from collections.abc import Generator
from typing import Iterable, AsyncIterator, List, Any

from typer import echo

from spinta import commands
from spinta.backends import Backend
from spinta.cli.helpers.data import read_model_data
from spinta.cli.helpers.errors import cli_error
from spinta.cli.helpers.export.components import CounterManager
from spinta.commands import build_data_patch_for_export
from spinta.components import Context, Model, pagination_enabled, DataItem, Action
from spinta.core.enums import Access
from spinta.formats.components import Format
from spinta.ufuncs.querybuilder.components import QueryParams
from spinta.utils.schema import NA


def validate_and_return_shallow_backend(context: Context, type_: str) -> Backend:
    """
    Validates and returns shallow backend.
    Shallow backend, is not loaded and is only used for type checking
    """
    config = context.get("config")
    backends = config.components['backends']
    if type_ not in backends:
        cli_error(
            f"Unknown output backend {type_!r}, from available: {list(backends.keys())}"
        )

    backend = config.components['backends'][type_]()
    backend.type = type_
    return backend


def validate_and_return_formatter(context: Context, type_: str) -> Format:
    config = context.get("config")
    exporters = config.exporters

    if type_ not in exporters:
        cli_error(
            f"Unknown output format {type_!r}, from available: {list(exporters.keys())}"
        )

    fmt = config.exporters[type_]
    return fmt


def get_data(context: Context, model: Model, access: Access):
    params = QueryParams()
    params.push = True

    page = None
    if pagination_enabled(model):
        page = commands.create_page(model.page)

    select_expr = None
    # Potentially we can also filter out all data with query
    # for now we only support `postgresql` output, which uses `prepare_export_patch`, that also does access filtering
    # in case we cannot do this with other outputs, we can also apply filtering with data fetch, but need to
    # make sure it all works together and all required data is fetched (for that reason currently we only use
    # `prepare_export_patch` to filter access data).

    # select_expr = Expr('select', Expr('bind', '_id'), *[
    #     Expr('bind', key) for key, prop in model.flatprops.items()
    #     if prop.access >= access and prop.name not in RESERVED_PROPERTY_NAMES
    # ])

    rows = read_model_data(
        context,
        model,
        params=params,
        page=page,
        query=select_expr
    )
    yield from rows


def create_data_items(model: Model, backend: Backend, data: Iterable):
    for item in data:
        yield DataItem(
            model=model,
            backend=backend,
            action=Action.INSERT,
            payload=item
        )


async def prepare_data_without_checks(
    context: Context,
    dstream: AsyncIterator[DataItem],
):
    async for data in dstream:
        data.payload = commands.rename_metadata(context, data.payload)
        data.given = commands.load(context, data.model, data.payload)
        yield data


async def prepare_export_patch(
    context: Context,
    dstream: AsyncIterator[DataItem],
    access: Access,
) -> AsyncIterator[DataItem]:
    # Truncated patch generation, fit for export command (fewer checks)
    # Main reason we use patch, is because `prepare_data_for_write` converts given data to backend specific data
    # It uses DataItem.patch to do it, so we need to generate one that also applies `export` command requirements.

    async for data in dstream:
        data.patch = build_data_patch_for_export(
            context,
            data.model or data.prop,
            given=data.given,
            access=access
        )
        if data.patch is NA:
            data.patch = {}

        if '_id' in data.given:
            data.patch['_id'] = data.given['_id']
        else:
            data.patch['_id'] = commands.gen_object_id(
                context, data.model.backend, data.model
            )

        data.patch['_revision'] = commands.gen_object_id(
            context, data.model.backend, data.model)
        yield data


async def export_data(
    context: Context,
    models: List[Model],
    fmt: Any,
    output: str,
    counter: CounterManager,
    access: Access
):
    txn = commands.gen_object_id(context, fmt)
    for model in models:
        data = get_data(context, model, access)
        await commands.export_data(
            context,
            model,
            fmt,
            data=data,
            path=output,
            counter=counter,
            txn=txn,
            access=access
        )
        counter.close_model(model)


def filter_models_by_access_verbose(models: Generator[Model], access: Access) -> Generator[Model]:
    for model in models:
        if model.access < access:
            echo(f"Skipping '{model.model_type()}' model, access level ('{model.access.name}') is lower than required ('{access.name}')")
            continue
        yield model



from __future__ import annotations

from typing import List

from spinta import commands
from spinta.components import Context, Model, Property
from spinta.exceptions import ModelReferenceNotFound, ModelReferenceKeyNotFound, \
    InvalidIntermediateTableMappingRefCount, UnableToMapIntermediateTable, SameModelIntermediateTableMapping
from spinta.types.datatype import Array, PartialArray, ArrayBackRef, Ref, DataType
from spinta.types.helpers import set_dtype_backend


@commands.link.register(Context, Array)
def link(context: Context, dtype: Array) -> None:
    set_dtype_backend(dtype)
    # In case of a dynamic array, dtype of items is not known.
    if dtype.items:
        commands.link(context, dtype.items.dtype)

    intermediate_model: str = dtype.model
    if intermediate_model is None:
        return

    if intermediate_model == dtype.prop.model.name:
        # Self reference.
        dtype.model = dtype.prop.model
    else:
        if not commands.has_model(context, dtype.prop.model.manifest, intermediate_model):
            raise ModelReferenceNotFound(dtype, ref=intermediate_model)
        dtype.model = commands.get_model(context, dtype.prop.model.manifest, intermediate_model)

    intermediate_model: Model = dtype.model
    if dtype.refprops:
        refprops = []
        raw_refprops: List[str] = dtype.refprops
        for rprop in raw_refprops:
            if rprop not in dtype.model.properties:
                raise ModelReferenceKeyNotFound(dtype, ref=rprop, model=dtype.model)
            refprops.append(dtype.model.properties[rprop])
        dtype.refprops = refprops
        dtype.explicit = True
    else:
        dtype.refprops = []

    dtype.left_prop, dtype.right_prop = _extract_intermediate_table_properties(dtype, intermediate_model)


def _compare_models(source: Model, target: str | Model) -> bool:
    if isinstance(target, Model):
        return source == target

    return source.name == target


def _extract_intermediate_table_properties(source: Array, intermediate_model: Model) -> (Property, Property):
    if source.refprops:
        if len(source.refprops) != 2:
            raise InvalidIntermediateTableMappingRefCount(source, ref_count=len(source.refprops))

        return source.refprops

    if intermediate_model.external and not intermediate_model.external.unknown_primary_key:
        if (
            len(intermediate_model.external.pkeys) == 2 and
            all(isinstance(prop.dtype, Ref) for prop in intermediate_model.external.pkeys)
        ):
            return intermediate_model.external.pkeys

    ref_properties = [prop for prop in intermediate_model.flatprops.values() if isinstance(prop.dtype, Ref)]
    if len(ref_properties) != 2:
        raise UnableToMapIntermediateTable(source)

    if all(_compare_models(source.prop.model, prop.dtype.model) for prop in ref_properties):
        raise SameModelIntermediateTableMapping(source)
    return ref_properties


@commands.link.register(Context, PartialArray)
def link(context: Context, dtype: PartialArray):
    dtype.prop.given.name = ''
    super_ = commands.link[Context, Array]
    return super_(context, dtype)


@commands.link.register(Context, ArrayBackRef)
def link(context: Context, dtype: ArrayBackRef):
    dtype.prop.given.name = ''

    set_dtype_backend(dtype)
    # In case of a dynamic array, dtype of items is not known.
    if dtype.items:
        commands.link(context, dtype.items.dtype)



import collections
import re
import uuid
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import TypedDict
from typing import overload

from starlette.requests import Request
from starlette.responses import Response
from toposort import toposort

from spinta import commands
from spinta import exceptions
from spinta.auth import authorized
from spinta.backends.constants import BackendFeatures
from spinta.backends.nobackend.components import NoBackend
from spinta.components import Action
from spinta.components import Config
from spinta.components import Context
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Node
from spinta.components import UrlParams
from spinta.exceptions import InvalidName
from spinta.manifests.components import Manifest
from spinta.nodes import load_node
from spinta.types.datatype import Ref, Array

namespace_is_lowercase = re.compile(r'^([a-z][a-z0-9]*)+(/[a-z][a-z0-9]*)+|([a-z][a-z0-9]*)$')

RESERVED_NAMES = {
    '_schema',
}


class NamespaceData(TypedDict):
    title: str
    description: str


def load_namespace_from_name(
    context: Context,
    manifest: Manifest,
    path: str,
    *,
    # Drop last element from path which is usually a model name.
    drop: bool = True,
) -> Namespace:
    ns: Optional[Namespace] = None
    parts: List[str] = []
    parts_ = [p for p in path.split('/') if p]
    if drop:
        parts_ = parts_[:-1]
    for part in [''] + parts_:
        parts.append(part)
        name = '/'.join(parts[1:])
        if not commands.has_namespace(context, manifest, name, loaded=True):
            ns = Namespace()
            data = {
                'type': 'ns',
                'name': name,
                'title': '',
                'description': '',
            }
            commands.load(context, ns, data, manifest)
            ns.generated = True
            commands.link(context, ns)
        else:
            ns = commands.get_namespace(context, manifest, name)
            pass

    return ns


@commands.load.register(Context, Namespace, dict, Manifest)
def load(
    context: Context,
    ns: Namespace,
    data: Dict[str, Any],
    manifest: Manifest,
    *,
    source: Manifest = None,
):
    load_node(context, ns, data)
    ns.path = manifest.path
    ns.manifest = manifest
    ns.access = manifest.access
    ns.backend = None
    ns.names = {}
    ns.models = {}
    commands.set_namespace(context, manifest, ns.name, ns)


@commands.link.register(Context, Namespace)
def link(context: Context, ns: Namespace):
    split = ns.name.split('/')
    if len(split) > 1:
        parent_ns = commands.get_namespace(context, ns.manifest, '/'.join(split[:-1]))
        if parent_ns:
            ns.parent = parent_ns
    elif ns.name != '':
        if commands.has_namespace(context, ns.manifest, ''):
            ns.parent = commands.get_namespace(context, ns.manifest, '')
    else:
        ns.parent = ns.manifest

    if isinstance(ns.parent, Namespace):
        ns.parent.names[ns.name] = ns


@commands.check.register(Context, Namespace)
def check(context: Context, ns: Namespace):
    config: Config = context.get('config')

    if config.check_names:
        name = ns.name
        if name and name not in RESERVED_NAMES:
            if namespace_is_lowercase.match(name) is None:
                raise InvalidName(ns, name=name, type='namespace')


@commands.authorize.register(Context, Action, Namespace)
def authorize(context: Context, action: Action, ns: Namespace):
    commands.authorize(context, action, ns, ns.manifest)


@commands.getall.register(Context, Namespace, Request)
async def getall(
    context: Context,
    ns: Namespace,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
    **kwargs
) -> Response:
    config: Config = context.get('config')
    if config.root and ns.is_root():
        ns = commands.get_namespace(context, ns.manifest, config.root)

    commands.authorize(context, action, ns)

    accesslog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        ns=ns.name,
        action=action.value,
    )
    return commands.getall(context, ns, request, ns.manifest, action=action, params=params)


@commands.getall.register(Context, Namespace)
def getall(
    context: Context,
    ns: Namespace,
    *,
    action: Optional[Action] = None,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
    **kwargs
):
    return _query_data(context, ns, action, dataset_, resource, **kwargs)


def _query_data(
    context: Context,
    ns: Namespace,
    action: Action,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
    **kwargs,
):
    models = commands.traverse_ns_models(
        context,
        ns,
        ns.manifest,
        action,
        dataset_=dataset_,
        resource=resource,
        internal=True,
    )
    for model in models:
        yield from commands.getall(context, model, model.backend, **kwargs)


def check_if_model_has_backend_and_source(model: Model):
    return not isinstance(model.backend, NoBackend) and (model.external and model.external.name)


def _model_matches_params(
    context: Context,
    model: Model,
    action: Action,
    dataset_: Optional[str] = None,
    resource: Optional[str] = None,
    internal: bool = False,
):

    if not internal and model.name.startswith('_'):
        return False

    if not authorized(context, model, action):
        return False

    if (
        dataset_ is None and
        resource is None
    ):
        return True

    if model.external is None:
        return False

    if (
        dataset_ is not None and (
            model.external.dataset is None or
            model.external.dataset.name != dataset_
        )
    ):
        return False

    if (
        resource is not None and (
            model.external.resource is None or
            model.external.resource.name != resource
        )
    ):
        return False

    return True


@commands.getone.register(Context, Request, Namespace)
async def getone(
    context: Context,
    request: Request,
    ns: Namespace,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    raise exceptions.ModelNotFound(model=ns.name)


@overload
@commands.in_namespace.register()  # noqa
def in_namespace(node: Node, ns: Namespace) -> bool:
    return node.name.startswith(ns.name)


@commands.in_namespace.register()  # noqa
def in_namespace(node: Node, parent: Node) -> bool:  # noqa
    while node:
        if node is parent:
            return True
        node = node.parent
    return False


@commands.wipe.register(Context, Namespace, type(None))
def wipe(context: Context, ns: Namespace, backend: type(None)):
    commands.authorize(context, Action.WIPE, ns)
    models = commands.traverse_ns_models(context, ns, ns.manifest, Action.WIPE, internal=True)
    models = sort_models_by_refs(models)
    for model in models:
        if BackendFeatures.WRITE in model.backend.features:
            commands.wipe(context, model, model.backend)


def sort_models_by_refs(models: Iterable[Model]) -> Iterator[Model]:
    models = {model.model_type(): model for model in models}
    graph = collections.defaultdict(set)
    for name, model in models.items():
        if model.base is None:
            graph[''].add(name)
        if model.base:
            base_model = model.base
            while base_model:
                graph[base_model.parent.name].add(name)
                if base_model.parent.base:
                    base_model = base_model.parent.base
                else:
                    base_model = None
        for dtype in iter_model_refs(model):
            ref = dtype.model.model_type()
            if ref in models:
                graph[ref].add(name)
    graph = toposort(graph)
    seen = {''}
    for group in graph:
        for name in sorted(group):
            if name in seen:
                continue
            seen.add(name)
            yield models[name]


def sort_models_by_base(models: Iterable[Model]) -> Iterator[Model]:
    models = {model.model_type(): model for model in models}
    graph = collections.defaultdict(set)
    for name, model in models.items():
        graph[''].add(name)
        for base in iter_model_base(model):
            ref = base.model_type()
            if ref in models:
                graph[ref].add(name)
    graph = toposort(graph)
    seen = {''}
    for group in graph:
        for name in sorted(group):
            if name in seen:
                continue
            seen.add(name)
            yield models[name]


def _topological_sort(model: Model, visited: dict, stack: list, dependencies: dict):
    visited[model.model_type()] = True
    for neighbor in dependencies[model.model_type()]:
        if not visited[neighbor.model_type()]:
            _topological_sort(neighbor, visited, stack, dependencies)
    stack.append(model)


def sort_models_by_ref_and_base(models: List[Model]):
    dependencies = collections.defaultdict(list)
    for model in models:
        if model.base and commands.identifiable(model.base):
            if model.base.parent in models:
                dependencies[model.model_type()].append(model.base.parent)
        for ref in iter_model_refs(model):
            if commands.identifiable(ref.prop):
                if ref.model in models:
                    dependencies[model.model_type()].append(ref.model)

    visited = {model.model_type(): False for model in models}
    stack = []

    for model in models:
        if not visited[model.model_type()]:
            _topological_sort(model, visited, stack, dependencies)

    return stack


def iter_model_base(model: Model) -> Iterator[Model]:
    if model.base:
        yield from iter_model_base(model.base.parent)
        yield model.base.parent


def iter_model_refs(model: Model) -> Iterator[Ref]:
    for prop in model.flatprops.values():
        if isinstance(prop.dtype, Array):
            prop = prop.dtype.items
        if isinstance(prop.dtype, Ref):
            yield prop.dtype




import contextlib

from spinta.datasets.components import ExternalBackend


class DaskBackend(ExternalBackend):
    type: str = 'dask'

    query_builder_type = 'dask'

    @contextlib.contextmanager
    def begin(self):
        yield




from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest, get_manifest_object_names


@commands.check.register(Context, Manifest)
def check(context: Context, manifest: Manifest):
    for node in get_manifest_object_names():
        for obj in commands.get_nodes(context, manifest, node).values():
            check(context, obj)



from __future__ import annotations

import enum
from dataclasses import dataclass

from spinta.components import Context
from spinta.core.enums import Access
from spinta.manifests.yaml.components import InlineManifest
from spinta.types.datatype import Ref, BackRef, ArrayBackRef, Inherit, PartialArray
from spinta.utils.naming import to_model_name


class MermaidClassDiagram:
    title: str | None
    classes: list[MermaidClass]
    relationships: list[MermaidRelationship]

    def __init__(self, title: str = None):
        self.title = title
        self.classes = []
        self.relationships = []

    def __str__(self):
        text = ""

        if self.title:
            text += f"---\n{self.title}\n---\n"

        text += "classDiagram\n"

        for clas in self.classes:
            text += f"{str(clas)}\n"

        for relationship in self.relationships:
            text += f"{str(relationship)}\n"

        return text

    def add_class(self, clas: MermaidClass) -> None:
        self.classes.append(clas)

    def add_relationship(self, relationship: MermaidRelationship) -> None:
        self.relationships.append(relationship)


class MermaidClass:
    name: str
    properties: list[MermaidProperty]

    def __init__(self, name: str, is_enum=False):
        self.name = name
        self.properties = []
        self.is_enum = is_enum

    def add_property(self, prop: MermaidProperty) -> None:
        self.properties.append(prop)

    def __str__(self):
        class_text = f"class {self.name} {{\n"
        if self.is_enum:
            class_text = f"{class_text}<<enumeration>>\n"
        properties_text = "".join(f"{str(prop)}\n" for prop in self.properties)
        class_text += properties_text
        class_text += "}"
        return class_text


@dataclass
class MermaidProperty:
    name: str
    access: Access | None = None
    type: str | None = None
    cardinality: bool | None = None
    multiplicity: str | None = None

    ACCESS_MAPPING = {
        Access.open: '+',
        Access.public: '#',
        Access.protected: '~',
        Access.private: '-'
    }

    def __str__(self):

        property_string = self.name

        if self.access is not None:
            access = self.ACCESS_MAPPING[self.access]
            property_string = f"{access} {property_string}"

        if self.type is not None:
            property_string = f"{property_string} : {self.type}"

        if self.cardinality or self.multiplicity:
            property_string = f"{property_string} [{int(self.cardinality) if self.cardinality is not None else ''}..{self.multiplicity if self.multiplicity is not None else ''}]"

        return property_string


class RelationshipDirection(enum.Enum):
    FORWARD = 'forward'
    BACKWARD = 'backward'


class RelationshipType(enum.Enum):
    ASSOCIATION = 'association'
    DEPENDENCY = 'dependency'
    INHERITANCE = 'inheritance'


@dataclass
class MermaidRelationship:
    node1: str
    node2: str
    type: RelationshipType
    direction: RelationshipDirection = RelationshipDirection.FORWARD
    cardinality: bool = False
    multiplicity: str = ""
    label: str = ""

    def __str__(self):

        if self.direction == RelationshipDirection.FORWARD:
            if self.type == RelationshipType.ASSOCIATION:
                arrow = '-->'
            if self.type == RelationshipType.DEPENDENCY:
                arrow = '..>'
            if self.type == RelationshipType.INHERITANCE:
                arrow = '--|>'
        else:
            if self.type == RelationshipType.ASSOCIATION:
                arrow = '<--'
            if self.type == RelationshipType.DEPENDENCY:
                arrow = '<..'
            if self.type == RelationshipType.INHERITANCE:
                arrow = '<|--'

        if self.cardinality or self.multiplicity:
            cardinality_multiplicity = f' "[{int(self.cardinality)}..{self.multiplicity}]"'
        else:
            cardinality_multiplicity = ""

        relationship_text = f"{self.node1} {arrow}{cardinality_multiplicity} {self.node2}"
        if self.label:
            relationship_text += f" : {self.label}"

        return relationship_text


def write_mermaid_manifest(context: Context, output: str, manifest: InlineManifest):

    mermaids = []

    for dataset_name, dataset in manifest.get_objects()["dataset"].items():
        mermaid = MermaidClassDiagram(title=dataset_name)

        models = manifest.get_objects()["model"]
        for model in models.values():
            model_dataset_name = model.external.dataset.name
            if model_dataset_name == dataset_name:
                mermaid_class = MermaidClass(name=model.basename)
                for model_property in model.get_given_properties().values():

                    if model_property.enum:
                        enum_class = MermaidClass(name=f"{model.basename}{to_model_name(model_property.name)}", is_enum=True)
                        for enum_property in model_property.enum:
                            enum_property = MermaidProperty(name=enum_property.strip('"'))
                            enum_class.add_property(enum_property)
                        mermaid.add_class(enum_class)
                        mermaid.add_relationship(
                            MermaidRelationship(
                                node1=mermaid_class.name,
                                node2=enum_class.name,
                                cardinality=True,
                                multiplicity="1",
                                type=RelationshipType.DEPENDENCY,
                                label=model_property.name
                            ))
                        
                    elif (
                        isinstance(model_property.dtype, Ref)
                        or isinstance(model_property.dtype, BackRef)
                    ):
                        multiplicity = '1'
                        mermaid.add_relationship(
                            MermaidRelationship(
                                node1=mermaid_class.name,
                                node2=model_property.dtype.model.basename,
                                cardinality=model_property.dtype.required,
                                multiplicity=multiplicity,
                                type=RelationshipType.ASSOCIATION,
                                label=model_property.name
                            )
                        )

                    # If it's a property that already is in a `base` model, we don't add it
                    elif isinstance(model_property.dtype, Inherit):
                        continue

                    elif (
                        isinstance(model_property.dtype, PartialArray)
                        or isinstance(model_property.dtype, ArrayBackRef)
                    ):
                        multiplicity = '*'

                        if hasattr(model_property.dtype, "items"):
                            if hasattr(model_property.dtype.items.dtype, "model"):
                                mermaid.add_relationship(
                                    MermaidRelationship(
                                        node1=mermaid_class.name,
                                        node2=model_property.dtype.items.dtype.model.basename,
                                        cardinality=model_property.dtype.items.dtype.required,
                                        multiplicity=multiplicity,
                                        type=RelationshipType.ASSOCIATION,
                                        label=model_property.name
                                    )
                                )
                            else:
                                mermaid_property = MermaidProperty(
                                    name=model_property.name,
                                    access=model_property.access,
                                    type=model_property.dtype.items.dtype.name,
                                    cardinality=model_property.dtype.items.dtype.required,
                                    multiplicity=multiplicity,
                                )
                                mermaid_class.add_property(mermaid_property)
                        else:

                            mermaid_property = MermaidProperty(
                                name=model_property.name,
                                access=model_property.access,
                                type=model_property.dtype.name,
                                cardinality=model_property.dtype.required,
                                multiplicity=multiplicity,
                            )
                            mermaid_class.add_property(mermaid_property)

                    else:

                        multiplicity = '1'

                        mermaid_property = MermaidProperty(
                            name=model_property.name,
                            access=model_property.access,
                            type=model_property.dtype.name,
                            cardinality=model_property.dtype.required,
                            multiplicity=multiplicity,
                        )
                        mermaid_class.add_property(mermaid_property)

                mermaid.add_class(mermaid_class)

                if model.base:
                    labels = []
                    for pk_property in model.base.pk:
                        labels.append(pk_property.name)
                    label = ", ".join(labels)
                    mermaid.add_relationship(
                        MermaidRelationship(
                            node1=mermaid_class.name,
                            node2=model.base.basename,
                            type=RelationshipType.INHERITANCE,
                            label=label
                        )
                    )

        mermaids.append(mermaid)

    with open(output, 'w') as file:

        for mermaid in mermaids:

            file.write(str(mermaid))



import numbers
from uuid import UUID

_LITERAL_TYPES = (
    numbers.Number, type(None), str
)


def is_str_uuid(value: str) -> bool:
    try:
        uuid_obj = UUID(value, version=4)
    except Exception:
        return False
    return str(uuid_obj) == value


def is_value_literal(value: object) -> bool:
    return isinstance(value, _LITERAL_TYPES)



from spinta import commands
from spinta.components import Context
from spinta.exceptions import IntermediateTableMappingInvalidType, IntermediateTableValueTypeMissmatch, \
    IntermediateTableRefModelMissmatch, IntermediateTableRefPropertyModelMissmatch, \
    IntermediateTableMissingMappingProperty
from spinta.types.datatype import Array, Ref


@commands.check.register(Context, Array)
def check(context: Context, dtype: Array):
    if dtype.items is not None:
        commands.check(context, dtype.items)

    if dtype.model is not None:
        if dtype.left_prop is None:
            raise IntermediateTableMissingMappingProperty(dtype, side="left")

        if dtype.right_prop is None:
            raise IntermediateTableMissingMappingProperty(dtype, side="right")

        if not isinstance(dtype.left_prop.dtype, Ref):
            raise IntermediateTableMappingInvalidType(
                dtype,
                property_name=dtype.left_prop.name,
                property_type=dtype.left_prop.dtype.name
            )

        if not isinstance(dtype.right_prop.dtype, Ref):
            raise IntermediateTableMappingInvalidType(
                dtype,
                property_name=dtype.right_prop.name,
                property_type=dtype.right_prop.dtype.name
            )

        if dtype.items is not None and not isinstance(dtype.items.dtype, type(dtype.right_prop.dtype)):
            raise IntermediateTableValueTypeMissmatch(
                dtype,
                array_type=dtype.items.dtype.name,
                intermediate_type=dtype.right_prop.dtype.name
            )

        if dtype.left_prop.dtype.model != dtype.prop.model:
            raise IntermediateTableRefModelMissmatch(
                dtype,
                array_model=dtype.prop.model.name,
                left_model=dtype.left_prop.dtype.model.name
            )

        if dtype.items.dtype.model != dtype.right_prop.dtype.model:
            raise IntermediateTableRefPropertyModelMissmatch(
                dtype,
                array_ref_model=dtype.items.dtype.model.name,
                right_model=dtype.right_prop.dtype.model.name
            )



from __future__ import annotations

from spinta.datasets.backends.sql.backends.postgresql.components import PostgreSQL
from spinta.datasets.backends.sql.ufuncs.query.components import SqlQueryBuilder


class PostgreSQLQueryBuilder(SqlQueryBuilder):
    backend: PostgreSQL



from dataclasses import dataclass
from typing import TypedDict

import sqlalchemy as sa
from sqlalchemy.engine.base import Engine as SaEngine

from spinta.core.ufuncs import Env
from spinta.ufuncs.querybuilder.components import Selected
from spinta.utils.data import take
from spinta.utils.schema import NA


@dataclass
class Engine:
    dsn: str  # sqlalchemy engine dsn
    schema: str = None
    encoding: str = NA

    def create(self) -> SaEngine:
        return sa.create_engine(self.dsn, **take({
            'encoding': self.encoding,
        }))


class SqlResource(Env):
    dsn: str

    def init(self, dsn: str):
        return self(dsn=dsn)


class FileSelected(TypedDict):
    name: Selected      # File name
    content: Selected   # File content



from typing import List, Dict, Tuple

import sqlalchemy as sa
from alembic.migration import MigrationContext
from alembic.operations import Operations
from sqlalchemy.engine.reflection import Inspector

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.commands.migrate.constants import EXCLUDED_MODELS
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name
from spinta.backends.postgresql.helpers.migrate.actions import MigrationHandler
from spinta.backends.postgresql.helpers.migrate.migrate import drop_all_indexes_and_constraints, model_name_key, \
    MigratePostgresMeta
from spinta.backends.postgresql.helpers.name import get_pg_table_name, get_pg_column_name, \
    get_pg_foreign_key_name, get_pg_file_name, PG_NAMING_CONVENTION
from spinta.utils.sqlalchemy import get_metadata_naming_convention
from spinta.cli.helpers.migrate import MigrateRename, MigrateMeta
from spinta.commands import create_exception
from spinta.components import Context, Model
from spinta.datasets.inspect.helpers import zipitems
from spinta.manifests.components import Manifest
from spinta.types.datatype import Ref, File
from spinta.types.namespace import sort_models_by_ref_and_base
from spinta.utils.schema import NA


@commands.migrate.register(Context, Manifest, PostgreSQL, MigrateMeta)
def migrate(context: Context, manifest: Manifest, backend: PostgreSQL, migrate_meta: MigrateMeta):
    conn = context.get(f'transaction.{backend.name}')
    ctx = MigrationContext.configure(conn, opts={
        "as_sql": migrate_meta.plan,
        "literal_binds": migrate_meta.plan,
        "target_metadata": backend.schema
    })
    op = Operations(ctx)
    inspector = sa.inspect(conn)
    metadata = sa.MetaData(bind=conn, naming_convention=get_metadata_naming_convention(PG_NAMING_CONVENTION))
    metadata.reflect(
        only=_filter_reflect_datasets(
            inspector,
            migrate_meta.datasets
        )
    )

    handler = MigrationHandler()
    meta = MigratePostgresMeta(
        inspector=inspector,
        handler=handler,
        rename=migrate_meta.rename
    )

    models = commands.get_models(context, manifest)
    models, tables = _filter_models_and_tables(
        models=models,
        existing_tables=inspector.get_table_names(),
        filtered_datasets=migrate_meta.datasets,
        rename=migrate_meta.rename
    )

    sorted_models = sort_models_by_ref_and_base(list(models.values()))
    sorted_model_names = list([model.name for model in sorted_models])
    # Do reverse zip, to ensure that sorted models get selected first
    zipped_names = zipitems(
        sorted_model_names,
        tables,
        model_name_key
    )

    for zipped_name in zipped_names:
        for new_model_name, old_table_name in zipped_name:
            # Skip Changelog and File table migrations, because this is done in DataType migration section
            if old_table_name and any(
                value in old_table_name for value in (TableType.CHANGELOG.value, TableType.FILE.value)
            ):
                continue

            # Skip excluded tables
            if old_table_name and old_table_name in EXCLUDED_MODELS:
                continue

            old = NA
            if old_table_name:
                name = get_pg_table_name(migrate_meta.rename.get_old_table_name(old_table_name))
                old = metadata.tables[name]

            new = commands.get_model(context, manifest, new_model_name) if new_model_name else new_model_name
            commands.migrate(context, backend, meta, old, new)
    _clean_up_file_type(inspector, sorted_models, handler, migrate_meta.rename)

    try:
        # Handle autocommit migrations differently
        if migrate_meta.autocommit:
            # Recreate new connection, that auto commits
            with backend.engine.connect().execution_options(isolation_level="AUTOCOMMIT") as conn:
                ctx = MigrationContext.configure(conn, opts={
                    "as_sql": migrate_meta.plan,
                    "literal_binds": migrate_meta.plan,
                    "target_metadata": backend.schema
                })
                op = Operations(ctx)
                handler.run_migrations(op)
            return

        # If in plan mode, just run the migrations
        # ctx is already setup so that it would not execute the code
        if migrate_meta.plan:
            with ctx.begin_transaction():
                handler.run_migrations(op)
            return

        # Begin transaction or take one, if it has already been opened
        if ctx._in_connection_transaction():
            trx = ctx.connection._transaction
        else:
            trx = ctx.begin_transaction()

        try:
            handler.run_migrations(op)

            # You can add custom logic that you might want to execute after running migrations, but before commiting
            # transaction, like `InternalSqlManifest` can update its own manifest schema, so you might want to try update
            # it here, incase it fails, transaction will be reverted and no changes will be made.
            if migrate_meta.migration_extension is not None:
                migrate_meta.migration_extension()
            trx.commit()
        except Exception:
            trx.rollback()
            raise

    except sa.exc.OperationalError as error:
        exception = create_exception(manifest, error)
        raise exception


def _filter_reflect_datasets(
    inspector: Inspector,
    datasets: list
):
    if not datasets:
        return None

    all_tables = inspector.get_table_names()
    return [table for table in all_tables if any(table.startswith(dataset) for dataset in datasets)]


def _filter_models_and_tables(
    models: Dict[str, Model],
    existing_tables: List[str],
    filtered_datasets: List[str],
    rename: MigrateRename
) -> Tuple[Dict[str, Model], List[str]]:
    tables = []

    # Filter if only specific dataset can be changed
    if filtered_datasets:
        filtered_models = {}
        for key, model in models.items():
            if model.external and model.external.dataset and model.external.dataset.name in filtered_datasets:
                filtered_models[key] = model
        models = filtered_models

        filtered_names = []
        for table_name in existing_tables:
            for dataset_name in filtered_datasets:
                if table_name.startswith(f'{dataset_name}/'):
                    # Check if its model or another sub dataset
                    additional_check = table_name.replace(f'{dataset_name}/', '', 1)
                    if '/' not in additional_check:
                        filtered_names.append(table_name)
        existing_tables = filtered_names

    for table in existing_tables:
        # Do not apply `get_pg_table_name`, since this will be done later on while zipping with `model_name_key`
        name = rename.get_table_name(table)
        if name not in models.keys():
            name = table
        tables.append(name)

    return models, tables


def _handle_foreign_key_constraints(inspector: Inspector, models: List[Model], handler: MigrationHandler,
                                    rename: MigrateRename):
    existing_table_names = set(inspector.get_table_names())

    for model in models:
        source_name = get_table_name(model)
        source_table = get_pg_table_name(source_name)
        old_name = get_pg_table_name(rename.get_old_table_name(source_name))
        foreign_keys = inspector.get_foreign_keys(old_name) if old_name in existing_table_names else []

        # Handle Base _id foreign key constraints
        id_constraint = next(
            (constraint for constraint in foreign_keys if constraint.get("constrained_columns") == ["_id"]), None
        )
        if id_constraint is not None:
            foreign_keys.remove(id_constraint)

        if model.base and commands.identifiable(model.base):
            add_constraint = True
            referent_table = get_pg_table_name(get_table_name(model.base.parent))
            fk_name = get_pg_foreign_key_name(referent_table, "_id")
            if id_constraint is not None:
                if id_constraint["name"] == fk_name and id_constraint["referred_table"] == referent_table:
                    add_constraint = False
                else:
                    handler.add_action(ma.DropConstraintMigrationAction(
                        table_name=source_table,
                        constraint_name=id_constraint["name"]
                    ), True)

            if add_constraint:
                handler.add_action(
                    ma.CreateForeignKeyMigrationAction(
                        source_table=source_table,
                        referent_table=referent_table,
                        constraint_name=fk_name,
                        local_cols=["_id"],
                        remote_cols=["_id"]
                    ), True
                )
        else:
            if id_constraint is not None:
                handler.add_action(ma.DropConstraintMigrationAction(
                    table_name=source_table,
                    constraint_name=id_constraint["name"]
                ), True)

        # Handle Ref foreign key constraints
        required_ref_props = {}
        for prop in model.flatprops.values():
            if isinstance(prop.dtype, Ref):
                if not prop.level or prop.level > 3:
                    column_name = get_pg_column_name(f"{prop.place}._id")
                    name = get_pg_foreign_key_name(source_table, column_name)
                    required_ref_props[name] = {
                        "name": name,
                        "constrained_columns": [column_name],
                        "referred_table": get_pg_table_name(get_table_name(prop.dtype.model)),
                        "referred_columns": ["_id"]
                    }

        for foreign_key in foreign_keys:
            if foreign_key["name"] not in required_ref_props.keys():
                handler.add_action(ma.DropConstraintMigrationAction(
                    table_name=source_table,
                    constraint_name=foreign_key["name"]
                ), True)
                continue

            constraint = required_ref_props[foreign_key["name"]]
            if (
                foreign_key["constrained_columns"] == constraint["constrained_columns"] and
                foreign_key["referred_table"] == constraint["referred_table"] and
                foreign_key["referred_columns"] == constraint["referred_columns"]
            ):
                del required_ref_props[foreign_key["name"]]
            else:
                handler.add_action(ma.DropConstraintMigrationAction(
                    table_name=source_table,
                    constraint_name=foreign_key["name"]
                ), True)

        for prop in required_ref_props.values():
            handler.add_action(ma.CreateForeignKeyMigrationAction(
                source_table=source_table,
                referent_table=prop["referred_table"],
                constraint_name=prop["name"],
                local_cols=prop["constrained_columns"],
                remote_cols=prop["referred_columns"]
            ), True)


def _clean_up_file_type(inspector: Inspector, models: List[Model], handler: MigrationHandler, rename: MigrateRename):
    allowed_file_tables = []
    existing_tables = []
    for model in models:
        existing_tables.append(rename.get_old_table_name(model.name))
        for prop in model.properties.values():
            if isinstance(prop.dtype, File):
                old_table = rename.get_old_table_name(get_table_name(model))
                old_column = rename.get_old_column_name(old_table, get_column_name(prop))
                allowed_file_tables.append(
                    get_pg_file_name(old_table, old_column)
                )

    for table in inspector.get_table_names():
        if TableType.FILE.value in table:
            split = table.split(f'{TableType.FILE.value}/')
            if split[0] in existing_tables:
                if table not in allowed_file_tables and not split[1].startswith("__"):
                    new_name = get_pg_file_name(split[0], f'__{split[1]}')
                    if inspector.has_table(new_name):
                        handler.add_action(ma.DropTableMigrationAction(
                            table_name=new_name
                        ))
                    handler.add_action(ma.RenameTableMigrationAction(
                        old_table_name=table,
                        new_table_name=new_name
                    ))
                    drop_all_indexes_and_constraints(inspector, table, new_name, handler)



from copy import copy

from spinta import commands
from spinta.components import Context, Property, Namespace
from spinta.components import Model
from spinta.datasets.components import Dataset, Resource, ExternalBackend
from spinta.datasets.inspect.helpers import zipitems, _backend_dsn, _dataset_key, _merge_resources, \
    _filter_models_for_dataset, coalesce, _merge_prefixes, _model_key, _resource_key, _merge_model_properties, \
    _property_key
from spinta.dimensions.prefix.components import UriPrefix
from spinta.manifests.components import Manifest
from spinta.types.datatype import Ref, DataType, Array, Object, Denorm
from spinta.utils.naming import Deduplicator
from spinta.utils.schema import NA
from spinta.utils.schema import NotAvailable


@commands.merge.register(Context, Manifest, Manifest, Manifest, bool)
def merge(context: Context, manifest: Manifest, old: Manifest, new: Manifest, has_manifest_priority: bool) -> None:
    backends = zipitems(
        old.backends.values(),
        new.backends.values(),
        _backend_dsn
    )

    backend_deduplicator = Deduplicator("_{}")
    for backend in backends:
        for o, n in backend:
            if not (o and n):
                if o:
                    backend_deduplicator(o.name)
                if n:
                    name = backend_deduplicator(n.name)
                    n.name = name
            merge(context, manifest, o, n)
    datasets = zipitems(
        commands.get_datasets(context, old).values(),
        commands.get_datasets(context, new).values(),
        _dataset_key,
    )
    deduplicator = Deduplicator("{}")
    for ds in datasets:
        for o, n in ds:
            if o:
                deduplicator(o.name)
            elif n and not o:
                name = deduplicator(n.name)
                n.name = name
                n.ns.name = name
            commands.merge(context, manifest, o, n, has_manifest_priority)


@commands.merge.register(Context, Manifest, NotAvailable, ExternalBackend)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: ExternalBackend) -> None:
    manifest.backends[new.name] = new


@commands.merge.register(Context, Manifest, ExternalBackend, ExternalBackend)
def merge(context: Context, manifest: Manifest, old: ExternalBackend, new: ExternalBackend) -> None:
    manifest.backends[old.name] = old


@commands.merge.register(Context, Manifest, ExternalBackend, NotAvailable)
def merge(context: Context, manifest: Manifest, old: ExternalBackend, new: NotAvailable) -> None:
    manifest.backends[old.name] = old


@commands.merge.register(Context, Manifest, NotAvailable, Dataset, bool)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Dataset, has_manifest_priority: bool) -> None:
    commands.set_dataset(context, manifest, new.name, new)
    _merge_resources(context, manifest, old, new)

    dataset_models = _filter_models_for_dataset(context, new.manifest, new)
    deduplicator = Deduplicator()
    for model in dataset_models:
        model.name = deduplicator(model.name)
        merge(context, manifest, NA, model, has_manifest_priority)

    new.manifest = manifest


@commands.merge.register(Context, Manifest, Dataset, Dataset, bool)
def merge(context: Context, manifest: Manifest, old: Dataset, new: Dataset, has_manifest_priority: bool) -> None:
    old.id = coalesce(old.id, new.id)
    old.description = coalesce(old.description, new.description)
    old.lang = coalesce(old.lang, new.lang)
    old.name = coalesce(old.name, new.name)
    old.manifest = coalesce(old.manifest, new.manifest)
    old.website = coalesce(old.website, new.website)
    old.source = coalesce(old.source, new.source)
    old.given = coalesce(old.given, new.given)
    old.level = coalesce(old.level, new.level)
    old.access = coalesce(old.access, new.access)
    old.title = coalesce(old.title, new.title)
    old.resources = coalesce(old.resources, new.resources)
    old.prefixes = coalesce(old.prefixes, new.prefixes)
    if old.ns and new.ns and old.ns.name == new.ns.name:
        commands.merge(context, manifest, old.ns, new.ns)
    else:
        old.ns = coalesce(old.ns, new.ns)
    commands.set_dataset(context, manifest, old.name, old)
    _merge_prefixes(context, manifest, old, new)
    _merge_resources(context, manifest, old, new)

    dataset_models = _filter_models_for_dataset(context, manifest, old)
    models = zipitems(
        dataset_models,
        commands.get_models(context, new.manifest).values(),
        _model_key
    )
    resource_list = []
    for res in new.resources.values():
        resource_list.append(_resource_key(res))
    deduplicator = Deduplicator()
    for model in models:
        for om, nm in model:
            if om:
                deduplicator(om.name)
            if om and not nm:
                if om.external and om.external.resource:
                    if _resource_key(om.external.resource) in resource_list:
                        om.external.name = None
            if not om and nm:
                if nm.external:
                    nm.external.dataset = old
                    name = deduplicator(f"{nm.external.dataset.name}/{nm.basename}")
                    nm.name = name
                    nm.ns.name = nm.external.dataset.name
            merge(context, manifest, om, nm, has_manifest_priority)


@commands.merge.register(Context, Manifest, Dataset, NotAvailable, bool)
def merge(context: Context, manifest: Manifest, old: Dataset, new: NotAvailable, has_manifest_priority: bool) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, UriPrefix)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: UriPrefix) -> None:
    dataset = new.parent
    commands.get_dataset(context, manifest, dataset.name).prefixes[new.name] = new


@commands.merge.register(Context, Manifest, UriPrefix, UriPrefix)
def merge(context: Context, manifest: Manifest, old: UriPrefix, new: UriPrefix) -> None:
    old.type = coalesce(new.type, old.type)
    old.id = coalesce(new.id, old.id)
    old.uri = coalesce(new.uri, old.uri)
    old.title = coalesce(new.title, old.title)
    old.description = coalesce(new.description, old.description)

    old.name = coalesce(old.name, new.name)


@commands.merge.register(Context, Manifest, UriPrefix, NotAvailable)
def merge(context: Context, manifest: Manifest, old: UriPrefix, new: NotAvailable) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, Namespace)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Namespace) -> None:
    commands.set_namespace(context, manifest, new.name, new)


@commands.merge.register(Context, Manifest, Namespace, Namespace)
def merge(context: Context, manifest: Manifest, old: Namespace, new: Namespace) -> None:
    old.keymap = coalesce(new.keymap, old.keymap)
    old.backend = coalesce(new.backend, old.backend)
    old.parent = coalesce(new.parent, old.parent)
    old.description = coalesce(new.description, old.description)
    old.generated = coalesce(new.generated, old.generated)
    old.lang = coalesce(new.lang, old.lang)
    old.enums = coalesce(new.enums, old.enums)

    old.name = coalesce(old.name, new.name)
    old.given = coalesce(old.given, new.given)
    old.title = coalesce(old.title, new.title)
    old.models = coalesce(old.models, new.models)
    old.access = coalesce(old.access, new.access)


@commands.merge.register(Context, Manifest, Namespace, NotAvailable)
def merge(context: Context, manifest: Manifest, old: Namespace, new: NotAvailable) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, Resource)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Resource) -> None:
    commands.get_dataset(context, manifest, new.dataset.name).resources[new.name] = new


@commands.merge.register(Context, Manifest, Resource, Resource)
def merge(context: Context, manifest: Manifest, old: Resource, new: Resource) -> None:
    old.type = coalesce(new.type, old.type)
    old.description = coalesce(new.description, old.description)
    old.comments = coalesce(new.comments, old.comments)
    old.lang = coalesce(new.lang, old.lang)

    old.backend = coalesce(old.backend, new.backend)
    old.dataset = coalesce(old.dataset, new.dataset)
    old.models = old.models
    old.eid = coalesce(old.eid, new.eid)
    old.given = coalesce(old.given, new.given)
    old.access = coalesce(old.access, new.access)
    old.title = coalesce(old.title, new.title)
    old.name = coalesce(old.name, new.name)
    old.level = coalesce(old.level, new.level)
    old.external = old.external
    old.prepare = coalesce(old.prepare, new.prepare)


@commands.merge.register(Context, Manifest, Resource, NotAvailable)
def merge(context: Context, manifest: Manifest, old: Resource, new: NotAvailable) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, Model, bool)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Model, has_manifest_priority: bool) -> None:
    old = copy(new)
    old.external = copy(old.external)
    old_name = old.name
    if f'{old.ns.name}/{old.basename}' != old.name and old.ns.name:
        name = f'{old.ns.name}/{old.basename}'
        old.name = name
        new.name = name
    if old.external and old.external.resource:
        resources = zipitems(
            old.external.dataset.resources.values(),
            [old.external.resource],
            _resource_key
        )
        for res in resources:
            for old_res, new_res in res:
                if old_res and new_res:
                    old.external.resource = old_res
                    if old_name != old.name:
                        if old_name in old_res.models:
                            del old_res.models[old_name]
                    old_res.models[old.name] = old
    old.manifest = manifest
    commands.set_model(context, manifest, old.name, old)
    _merge_model_properties(context, manifest, old, new, has_manifest_priority)


@commands.merge.register(Context, Manifest, Model, Model, bool)
def merge(context: Context, manifest: Manifest, old: Model, new: Model, has_manifest_priority: bool) -> None:
    old.description = coalesce(new.description, old.description)
    old.comments = coalesce(new.comments, old.comments)
    old.lang = coalesce(new.lang, old.lang)

    old.id = coalesce(old.id, new.id)
    old.eid = coalesce(old.eid, new.eid)
    old.base = coalesce(old.base, new.base)
    old.given = coalesce(old.given, new.given)
    old.name = coalesce(old.name, new.name)
    old.title = coalesce(old.title, new.title)
    old.access = coalesce(old.access, new.access)
    old.level = coalesce(old.level, new.level)
    old.external = coalesce(old.external, new.external)
    old.manifest = manifest

    commands.set_model(context, manifest, old.name, old)
    _merge_model_properties(context, manifest, old, new, has_manifest_priority)

    if old.external and new.external:
        if not has_manifest_priority or (old.external.unknown_primary_key and not new.external.unknown_primary_key):
            keys = zipitems(
                old.properties.values(),
                new.external.pkeys,
                _property_key
            )
            new_keys = []
            for key in keys:
                for o, n in key:
                    if o and n:
                        new_keys.append(o)
            if new_keys:
                old.external.pkeys = new_keys
                old.external.unknown_primary_key = new.external.unknown_primary_key


@commands.merge.register(Context, Manifest, Model, NotAvailable, bool)
def merge(context: Context, manifest: Manifest, old: Model, new: NotAvailable, has_manifest_priority: bool) -> None:
    if old.external and not old.external.name:
        for prop in old.properties.values():
            if prop.external:
                prop.external.name = None


@commands.merge.register(Context, Manifest, NotAvailable, Property, bool)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Property, has_manifest_priority: bool) -> None:
    if new.external:
        new.external.prop = new
    new.dtype.prop = new
    merge(context, manifest, new.dtype, new.dtype)
    new.model.properties[new.name] = new


@commands.merge.register(Context, Manifest, Property, Property, bool)
def merge(context: Context, manifest: Manifest, old: Property, new: Property, has_manifest_priority: bool) -> None:
    if not has_manifest_priority:
        merged = old
        merged.type = coalesce(new.type, old.type)
        merged.uri = coalesce(new.uri, old.uri)
        merged.description = coalesce(new.description, old.description)
        merged.enum = coalesce(new.enum, old.enum)
        merged.enums = coalesce(new.enums, old.enums)
        merged.unit = coalesce(new.unit, old.unit)
        merged.comments = coalesce(new.comments, old.comments)
        merged.lang = coalesce(new.lang, old.lang)

        merged.model = coalesce(old.model, new.model)
        merged.external = coalesce(old.external, new.external)
        merged.place = coalesce(old.place, new.place)
        merged.name = coalesce(old.name, new.name)
        merged.given = coalesce(old.given, new.given)
        merged.title = coalesce(old.title, new.title)
        merged.level = coalesce(old.level, new.level)
        merged.access = coalesce(old.access, new.access)
        merged.dtype.prop = merged

        old.model.properties[old.name] = merged

        merge(context, manifest, merged.dtype, new.dtype)


@commands.merge.register(Context, Manifest, Property, NotAvailable, bool)
def merge(context: Context, manifest: Manifest, old: Property, new: NotAvailable, has_manifest_priority: bool) -> None:
    if old.external:
        old.external.name = None
    model = commands.get_model(context, manifest, old.model.name)
    model.properties[old.name] = old


@commands.merge.register(Context, Manifest, DataType, Array)
def merge(context: Context, manifest: Manifest, old: DataType, new: Array) -> None:
    merged = new
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)
    models = zipitems(
        [merged.items.model],
        commands.get_models(context, manifest).values(),
        _model_key
    )
    for model in models:
        for o, n in model:
            if o and n:
                properties = zipitems(
                    merged.items,
                    o.properties.values(),
                    _property_key
                )
                for prop in properties:
                    for po, pn in prop:
                        if po and pn:
                            merged.items = po
                            break
                    else:
                        continue
                    break
            break
        else:
            continue
        break

    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, Object)
def merge(context: Context, manifest: Manifest, old: DataType, new: Object) -> None:
    merged = new
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)

    new_properties = {}
    for key, value in merged.properties.items():
        new_key = key
        new_value = value
        models = zipitems(
            [value.model],
            commands.get_models(context, manifest).values(),
            _model_key
        )
        for model in models:
            for o, n in model:
                if o and n:
                    properties = zipitems(
                        merged.items,
                        o.properties.values(),
                        _property_key
                    )
                    for prop in properties:
                        for po, pn in prop:
                            if po and pn:
                                new_value = po
                                new_key = po.name
                                break
                        else:
                            continue
                        break
                break
            else:
                continue
            break
        new_properties[new_key] = new_value
    merged.properties = new_properties

    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, Ref)
def merge(context: Context, manifest: Manifest, old: DataType, new: Ref) -> None:
    merged = copy(new)
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)

    models = zipitems(
        [merged.model],
        commands.get_models(context, manifest).values(),
        _model_key
    )
    for model in models:
        for o, n in model:
            if o and n:
                if n.external and merged.prop.model.external:
                    if n.external.dataset is merged.prop.model.external.dataset:
                        merged.model = n
                        if not n.external.unknown_primary_key and n.external.pkeys:
                            merged.refprops = n.external.pkeys
                        else:
                            new_refprops = []
                            properties = zipitems(
                                merged.refprops,
                                n.properties.values(),
                                _property_key
                            )
                            for prop in properties:
                                for po, pn in prop:
                                    if po and pn:
                                        new_refprops.append(pn)

                            if not new_refprops:
                                new_refprops = merged.refprops
                            merged.refprops = new_refprops
                        break
        else:
            continue
        break
    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, Denorm)
def merge(context: Context, manifest: Manifest, old: DataType, new: Denorm) -> None:
    merged = new
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)

    models = zipitems(
        [merged.rel_prop.model],
        commands.get_models(context, manifest).values(),
        _model_key
    )
    for model in models:
        for o, n in model:
            if o and n:
                properties = zipitems(
                    merged.items,
                    o.properties.values(),
                    _property_key
                )
                for prop in properties:
                    for po, pn in prop:
                        if po and pn:
                            merged.rel_prop = po
                            break
                    else:
                        continue
                    break
            break
        else:
            continue
        break

    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, DataType)
def merge(context: Context, manifest: Manifest, old: DataType, new: DataType) -> None:
    merged = new
    merged.name = coalesce(new.name, old.name)
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)
    old.prop.dtype = merged



from typing import TypedDict, Callable, Dict

from spinta import commands
from spinta.components import Namespace, Model, Node, Context
from spinta.datasets.components import Dataset
from spinta.exceptions import DatasetNotFound, NamespaceNotFound, ModelNotFound, ManifestObjectNotDefined, \
    NotImplementedFeature
from spinta.manifests.components import Manifest


class _FunctionTypes(TypedDict):
    has: Callable
    get: Callable
    set: Callable
    get_all: Callable


NODE_FUNCTION_MAPPER = {
    'model': _FunctionTypes(
        has=commands.has_model,
        get=commands.get_model,
        set=commands.set_model,
        get_all=commands.get_models
    ),
    'ns': _FunctionTypes(
        has=commands.has_namespace,
        get=commands.get_namespace,
        set=commands.set_namespace,
        get_all=commands.get_namespaces
    ),
    'dataset': _FunctionTypes(
        has=commands.has_dataset,
        get=commands.get_dataset,
        set=commands.set_dataset,
        get_all=commands.get_datasets
    )
}


@commands.has_node_type.register(Context, Manifest, str)
def has_object_type(context: Context, manifest: Manifest, obj_type: str, **kwargs):
    return obj_type in manifest.get_objects()


@commands.has_node.register(Context, Manifest, str, str)
def has_object(context: Context, manifest: Manifest, obj_type: str, obj: str, **kwargs):
    if obj_type in NODE_FUNCTION_MAPPER:
        return NODE_FUNCTION_MAPPER[obj_type]['has'](context, manifest, obj, **kwargs)
    raise ManifestObjectNotDefined(obj=obj_type)


@commands.get_node.register(Context, Manifest, str, str)
def get_node(context: Context, manifest: Manifest, obj_type: str, obj: str, **kwargs):
    if obj_type in NODE_FUNCTION_MAPPER:
        return NODE_FUNCTION_MAPPER[obj_type]['get'](context, manifest, obj, **kwargs)
    raise ManifestObjectNotDefined(obj=obj_type)


@commands.get_nodes.register(Context, Manifest, str)
def get_nodes(context: Context, manifest: Manifest, obj_type: str, **kwargs):
    if obj_type in NODE_FUNCTION_MAPPER:
        return NODE_FUNCTION_MAPPER[obj_type]['get_all'](context, manifest, **kwargs)
    raise ManifestObjectNotDefined(obj=obj_type)


@commands.set_node.register(Context, Manifest, str, str, Node)
def set_node(context: Context, manifest: Manifest, obj_type: str, obj_name, obj: Node, **kwargs):
    if obj_type in NODE_FUNCTION_MAPPER:
        return NODE_FUNCTION_MAPPER[obj_type]['set'](context, manifest, obj_name, obj, **kwargs)
    raise ManifestObjectNotDefined(obj=obj_type)


@commands.has_model.register(Context, Manifest, str)
def has_model(context: Context, manifest: Manifest, model: str, **kwargs):
    return model in manifest.get_objects()['model']


@commands.get_model.register(Context, Manifest, str)
def get_model(context: Context, manifest: Manifest, model: str, **kwargs):
    if has_model(context, manifest, model):
        return manifest.get_objects()['model'][model]
    raise ModelNotFound(model=model)


@commands.get_models.register(Context, Manifest)
def get_models(context: Context, manifest: Manifest, **kwargs):
    return manifest.get_objects()['model']


@commands.set_model.register(Context, Manifest, str, Model)
def set_model(context: Context, manifest: Manifest, model_name: str, model: Model, **kwargs):
    manifest.get_objects()['model'][model_name] = model


@commands.set_models.register(Context, Manifest, dict)
def set_models(context: Context, manifest: Manifest, models: Dict[str, Model], **kwargs):
    manifest.get_objects()['model'] = models


@commands.has_namespace.register(Context, Manifest, str)
def has_namespace(context: Context, manifest: Manifest, namespace: str, **kwargs):
    return namespace in manifest.get_objects()['ns']


@commands.get_namespace.register(Context, Manifest, str)
def get_namespace(context: Context, manifest: Manifest, namespace: str, **kwargs):
    if has_namespace(context, manifest, namespace):
        return manifest.get_objects()['ns'][namespace]
    raise NamespaceNotFound(namespace=namespace)


@commands.get_namespaces.register(Context, Manifest)
def get_namespaces(context: Context, manifest: Manifest, **kwargs):
    return manifest.get_objects()['ns']


@commands.set_namespace.register(Context, Manifest, str, Namespace)
def set_namespace(context: Context, manifest: Manifest, namespace: str, ns: Namespace, **kwargs):
    manifest.get_objects()['ns'][namespace] = ns


@commands.has_dataset.register(Context, Manifest, str)
def has_dataset(context: Context, manifest: Manifest, dataset: str, **kwargs):
    return dataset in manifest.get_objects()['dataset']


@commands.get_dataset.register(Context, Manifest, str)
def get_dataset(context: Context, manifest: Manifest, dataset: str, **kwargs):
    if has_dataset(context, manifest, dataset):
        return manifest.get_objects()['dataset'][dataset]
    raise DatasetNotFound(dataset=dataset)


@commands.get_datasets.register(Context, Manifest)
def get_datasets(context: Context, manifest: Manifest, **kwargs):
    return manifest.get_objects()['dataset']


@commands.set_dataset.register(Context, Manifest, str, Dataset)
def set_dataset(context: Context, manifest: Manifest, dataset_name: str, dataset: Dataset, **kwargs):
    manifest.get_objects()['dataset'][dataset_name] = dataset


@commands.get_dataset_models.register(Context, Manifest, str)
def get_dataset_models(context: Context, manifest: Manifest, dataset_name: str, **kwargs):
    dataset = commands.get_dataset(context, manifest, dataset_name)
    models = commands.get_models(context, manifest)
    filtered_list = {}
    for key, model in models.items():
        if model.external and model.external.dataset == dataset:
            filtered_list[key] = model
    return filtered_list


@commands.update_manifest_dataset_schema.register(Context, Manifest, Manifest)
def update_manifest_dataset_schema(context: Context, manifest: Manifest, target_manifest: Manifest, **kwargs):
    raise NotImplementedFeature(manifest, "Ability to modify non-dynamic manifest's schema")



from __future__ import annotations

import contextlib
from typing import Any, Type
from typing import Dict
from typing import Optional
from typing import Set

from spinta.backends.constants import BackendOrigin, BackendFeatures
from spinta.core.ufuncs import Env
from spinta.ufuncs.resultbuilder.components import ResultBuilder


class Backend:
    metadata = {
        'name': 'backend',
    }

    type: str
    name: str
    origin: BackendOrigin
    features: Set[BackendFeatures] = set()

    # Original configuration values given in manifest, this is used to restore
    # manifest back to its original form.
    config: Dict[str, Any]

    available: bool = True

    # Query builder's type in config.components['querybuilders']
    # by default, '' is QueryBuilder
    query_builder_type: str = ''
    # Later on type should be changed to `QueryBuilder`
    query_builder_class: Type[Env]

    # Result builder's type in config.components['resultbuilders']
    # by default, '' is ResultBuilder
    result_builder_type: str = ''
    result_builder_class: Type[ResultBuilder]

    def __repr__(self):
        return (
            f'<{self.__class__.__module__}.{self.__class__.__name__}'
            f'(name={self.name!r}) at 0x{id(self):02x}>'
        )

    @contextlib.contextmanager
    def transaction(self):
        raise NotImplementedError

    @contextlib.contextmanager
    def begin(self):
        raise NotImplementedError

    def bootstrapped(self):
        raise NotImplementedError

    # Checks if backend supports specific feature
    def supports(self, feature: BackendFeatures) -> bool:
        return feature in self.features


SelectTree = Optional[Dict[str, 'SelectTree']]



from spinta.components import UrlParams, Property
from spinta.core.ufuncs import Env
from spinta.ufuncs.propertyresolver.components import PropertyResolver


class RequestParamsBuilder(Env):
    params: UrlParams

    def init(self, params: UrlParams):
        return self(
            params=params,
            property_resolver=PropertyResolver(context=self.context).init(model=params.model)
        )

    def resolve_property(self, *args, **kwargs) -> Property:
        return self.property_resolver.resolve_property(*args, **kwargs)





from spinta import commands
from spinta.components import Context
from spinta.exceptions import RefPropTypeMissmatch
from spinta.types.datatype import ExternalRef, Ref


@commands.check.register(Context, Ref)
def check(context: Context, dtype: Ref):
    for prop in dtype.properties.values():
        commands.check(context, prop)


@commands.check.register(Context, ExternalRef)
def check(context: Context, dtype: ExternalRef):
    for prop in dtype.properties.values():
        commands.check(context, prop)
        for ref_prop in dtype.refprops:
            if ref_prop.name == prop.name:
                if not isinstance(ref_prop.dtype, type(prop.dtype)):
                    raise RefPropTypeMissmatch(prop.dtype, refprop=ref_prop.name, required_type=ref_prop.dtype.name, given_type=prop.dtype.name)



from typing import Optional, Union

from spinta import commands
from spinta.backends import Backend
from spinta.cli.helpers.errors import cli_error
from spinta.components import Context, Model, DataSubItem, Property
from spinta.core.enums import Access
from spinta.formats.components import Format
from spinta.types.datatype import DataType, Ref, ExternalRef, Denorm, Object, Inherit, BackRef, Array, File
from spinta.utils.data import take
from spinta.utils.nestedstruct import flatten_value
from spinta.utils.schema import NA, NotAvailable


@commands.export_data.register(Context, Model, Format)
def export_data(context: Context, model: Model, fmt: Format, *, data: object, **kwargs):
    cli_error(
        f"Export for {type(fmt)!r} Format is supported yet."
    )


@commands.export_data.register(Context, Model, Backend)
def export_data(context: Context, model: Model, backend: Backend, *, data: object, **kwargs):
    cli_error(
        f"Export for {backend.type!r} Backend is supported yet."
    )


@commands.before_export.register(Context, Model, Backend)
def before_export(
    context: Context,
    model: Model,
    backend: Backend,
    *,
    data: DataSubItem
):
    patch = take(['_id'], data.patch)
    patch['_revision'] = take('_revision', data.patch, data.saved)
    for prop in take(model.properties).values():
        if not prop.dtype.inherited:
            prop_data = data[prop.name]
            value = commands.before_export(
                context,
                prop.dtype,
                backend,
                data=prop_data,
            )
            patch.update(value)
    return patch


@commands.before_export.register(Context, Property, Backend)
def before_export(
    context: Context,
    prop: Property,
    backend: Backend,
    *,
    data: DataSubItem
):
    return before_export(context, prop.dtype, backend, data=data)


@commands.before_export.register(Context, DataType, Backend)
def before_export(
    context: Context,
    dtype: DataType,
    backend: Backend,
    *,
    data: DataSubItem
):
    return take(all, {dtype.prop.place: data.patch})


@commands.before_export.register(Context, Ref, Backend)
def before_export(
    context: Context,
    dtype: Ref,
    backend: Backend,
    *,
    data: DataSubItem,
) -> dict:
    # If patch is None, it means that parent was set to null, meaning all children should also be set to null

    if data.patch is None:
        patch = {}

        if not dtype.inherited:
            patch[f'{dtype.prop.place}._id'] = None

        for prop in dtype.properties.values():
            value = commands.before_export(
                context,
                prop.dtype,
                backend,
                data=data,
            )
            patch.update(value)
        return patch

    patch = flatten_value(data.patch, dtype.prop)
    return {
        f'{dtype.prop.place}.{k}': v for k, v in patch.items() if k
    }


@commands.before_export.register(Context, ExternalRef, Backend)
def before_export(
    context: Context,
    dtype: ExternalRef,
    backend: Backend,
    *,
    data: DataSubItem,
) -> dict:
    # If patch is None, it means that parent was set to null, meaning all children should also be set to null
    if data.patch is None:
        patch = {}
        if not dtype.inherited:
            if dtype.explicit or not dtype.model.external.unknown_primary_key:
                for ref_prop in dtype.refprops:
                    patch[f'{dtype.prop.place}.{ref_prop.name}'] = None
            else:
                patch[f'{dtype.prop.place}._id'] = None

        for prop in dtype.properties.values():
            value = commands.before_export(
                context,
                prop.dtype,
                backend,
                data=data,
            )
            patch.update(value)

        return patch

    patch = flatten_value(data.patch, dtype.prop)
    return {
        f'{dtype.prop.place}.{k}': v for k, v in patch.items() if k
    }


@commands.before_export.register(Context, Denorm, Backend)
def before_export(
    context: Context,
    dtype: Denorm,
    backend: Backend,
    *,
    data: DataSubItem,
) -> dict:
    # If patch is None, it means that all children should be set to null
    if data.patch is None:
        patch = commands.before_export(context, dtype.rel_prop.dtype, backend, data=data)
        return {
            key.replace(dtype.rel_prop.place, dtype.prop.place): value
            for key, value in patch.items() if key != dtype.rel_prop.place
        }

    patch = flatten_value(data.patch, dtype.prop)
    key = dtype.prop.place.split('.', maxsplit=1)[-1]
    if patch.get(key):
        return {dtype.prop.place: patch.get(key)}
    else:
        return {}


@commands.before_export.register(Context, Object, Backend)
def before_export(
    context: Context,
    dtype: Object,
    backend: Backend,
    *,
    data: DataSubItem,
) -> dict:
    patch = {}
    # If patch is None, then just set every child to None
    if data.patch is None:
        for prop in dtype.properties.values():
            value = commands.before_export(
                context,
                prop.dtype,
                backend,
                data=data,
            )
            patch.update(value)
        return patch

    for prop in dtype.properties.values():
        value = commands.before_export(
            context,
            prop.dtype,
            backend,
            data=data[prop.name],
        )
        patch.update(value)
    return patch


@commands.validate_export_output.register(Context, Backend, type(None))
def validate_export_output(context: Context, backend: Backend, output):
    cli_error(
        "Output argument is required (`--output`)."
    )


@commands.validate_export_output.register(Context, Format, type(None))
def validate_export_output(context: Context, backend: Format, output):
    cli_error(
        "Output argument is required (`--output`)."
    )


@commands.validate_export_output.register(Context, Backend, object)
def validate_export_output(context: Context, backend: Backend, output):
    return


@commands.validate_export_output.register(Context, Format, object)
def validate_export_output(context: Context, fmt: Format, output):
    return


@commands.build_data_patch_for_export.register(Context, Model)
def build_data_patch_for_export(
    context: Context,
    model: Model,
    *,
    given: dict,
    access: Access
) -> dict:
    props = take(model.properties).values()
    props = [prop for prop in props if prop.access >= access]

    patch = {}
    for prop in props:
        value = build_data_patch_for_export(
            context,
            prop.dtype,
            given=given.get(prop.name, NA),
            access=access
        )
        if value is not NA:
            patch[prop.name] = value
    return patch


@commands.build_data_patch_for_export.register(Context, Property)
def build_data_patch_for_export(
    context: Context,
    prop: Property,
    *,
    given: dict,
    access: Access
) -> dict:
    value = build_data_patch_for_export(
        context,
        prop.dtype,
        given=given.get(prop.name, NA),
        access=access
    )
    if value is not NA:
        return {prop.name: value}
    else:
        return {}


@commands.build_data_patch_for_export.register(Context, DataType)
def build_data_patch_for_export(
    context: Context,
    dtype: DataType,
    *,
    given: Optional[object],
    access: Access
) -> Union[dict, NotAvailable]:
    if given is NA:
        given = dtype.default
    return given


@commands.build_data_patch_for_export.register(Context, Object)
def build_data_patch_for_export(
    context: Context,
    dtype: Object,
    *,
    given: Optional[dict],
    access: Access
) -> Union[dict, NotAvailable]:
    props = take(dtype.properties).values()

    props = [prop for prop in props if prop.access >= access]

    patch = {}
    for prop in props:
        value = build_data_patch_for_export(
            context,
            prop.dtype,
            given=given.get(prop.name, NA) if given else NA,
            access=access
        )
        if value is not NA:
            patch[prop.name] = value
    return patch or NA


@commands.build_data_patch_for_export.register(Context, Ref)
def build_data_patch_for_export(
    context: Context,
    dtype: Ref,
    *,
    given: Optional[dict],
    access: Access
) -> Union[dict, NotAvailable]:
    if given is None:
        return given

    patch = {
        '_id': given.get('_id', NA)
    }

    props = take(dtype.properties).values()
    props = [prop for prop in props if prop.access >= access]

    for prop in props:
        value = build_data_patch_for_export(
            context,
            prop.dtype,
            given=given.get(prop.name, NA) if given else NA,
            access=access
        )
        if value is not NA:
            patch[prop.name] = value
    return patch or NA


@commands.build_data_patch_for_export.register(Context, ExternalRef)
def build_data_patch_for_export(
    context: Context,
    dtype: ExternalRef,
    *,
    given: Optional[dict],
    access: Access
) -> Union[dict, NotAvailable]:
    if given is None:
        return given

    patch = {}

    refprops = [prop for prop in dtype.refprops if prop.access >= access]
    for prop in refprops:
        value = build_data_patch_for_export(
            context,
            prop.dtype,
            given=given.get(prop.name, NA) if given else NA,
            access=access
        )
        if value is not NA:
            patch[prop.name] = value

    props = take(dtype.properties).values()
    props = [prop for prop in props if prop.access >= access]

    for prop in props:
        value = build_data_patch_for_export(
            context,
            prop.dtype,
            given=given.get(prop.name, NA) if given else NA,
            access=access
        )
        if value is not NA:
            patch[prop.name] = value
    return patch or NA


@commands.build_data_patch_for_export.register(Context, Inherit)
def build_data_patch_for_export(
    context: Context,
    dtype: Inherit,
    *,
    given: Optional[object],
    access: Access
) -> Union[dict, NotAvailable]:
    # Needs to be implemented when it's possible to modify Inherit type
    return NA


@commands.build_data_patch_for_export.register(Context, BackRef)
def build_data_patch_for_export(
    context: Context,
    dtype: BackRef,
    *,
    given: Optional[dict],
    access: Access
) -> Union[dict, NotAvailable]:
    return NA


@commands.build_data_patch_for_export.register(Context, Array)
def build_data_patch_for_export(
    context: Context,
    dtype: Array,
    *,
    given: Optional[Union[list, NotAvailable]],
    access: Access
) -> Union[list, None, NotAvailable]:
    if dtype.items.access < access:
        return NA

    if given is NA:
        return []

    if given is None:
        return []

    return [
        build_data_patch_for_export(
            context,
            dtype.items.dtype,
            given=value,
            access=access
        )
        for value in given
    ]


@commands.build_data_patch_for_export.register(Context, File)
def build_data_patch_for_export(
    context: Context,
    dtype: File,
    *,
    given: Optional[dict],
    access: Access
) -> Union[dict, NotAvailable]:
    given = {
        '_id': given.get('_id', None) if given else None,
        '_content_type': given.get('_content_type', None) if given else None,
        '_content': given.get('_content', NA) if given else NA,
    }
    given = {k: v for k, v in given.items() if v is not NA}
    return given or NA



import enum


class TableType(enum.Enum):
    MAIN = ''
    LIST = '/:list'
    CHANGELOG = '/:changelog'
    CACHE = '/:cache'
    FILE = '/:file'


class BackendOrigin(enum.Enum):
    """Origin where backend was defined.

    Backend can be defined in multiple places, for example backend can be
    defined in a configuration file or inline in manifest.
    """

    config = 'config'
    manifest = 'manifest'
    resource = 'resource'


class BackendFeatures(enum.Enum):
    # Files are stored in blocks and file metadata must include _bsize and
    # _blocks properties.
    FILE_BLOCKS = 'FILE_BLOCKS'

    # Backend supports write operations.
    WRITE = 'WRITE'

    # Backend supports pagination when reading data
    PAGINATION = 'PAGINATION'

    # Backend supports
    EXPAND = 'EXPAND'



from __future__ import annotations

from typing import Any
from typing import List
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import overload

import sqlalchemy as sa
from sqlalchemy.sql.functions import Function

from spinta.auth import authorized
from spinta.components import Action, Page
from spinta.components import Property
from spinta.core.ufuncs import Bind, GetAttr
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import Negative
from spinta.core.ufuncs import Unresolved
from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.sql.ufuncs.query.components import SqlQueryBuilder
from spinta.datasets.backends.sql.ufuncs.query.helpers import select_ref_foreign_key_properties, \
    select_external_ref_foreign_key_properties
from spinta.dimensions.enum.helpers import prepare_enum_value
from spinta.exceptions import PropertyNotFound, SourceCannotBeList, NoExternalName, NotImplementedFeature
from spinta.types.datatype import DataType, Denorm, Object, Array, ExternalRef
from spinta.types.datatype import PrimaryKey
from spinta.types.datatype import Ref
from spinta.types.datatype import String
from spinta.types.datatype import UUID
from spinta.types.text.components import Text
from spinta.types.text.helpers import determine_language_property_for_text
from spinta.ufuncs.components import ForeignProperty
from spinta.ufuncs.querybuilder.components import LiteralProperty, Selected, ReservedProperty
from spinta.ufuncs.querybuilder.helpers import get_language_column, process_literal_value
from spinta.ufuncs.querybuilder.ufuncs import Star
from spinta.utils.data import take
from spinta.utils.itertools import flatten, ensure_list
from spinta.utils.schema import NA


def _gather_selected_properties(env: SqlQueryBuilder):
    result = []
    if env.selected:
        for selected in env.selected.values():
            if selected and selected.prop:
                result.append(selected.prop)
    return result


@ufunc.resolver(SqlQueryBuilder, str, object, names=['eq', 'ne'])
def eq(env: SqlQueryBuilder, op: str, field: str, value: Any):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    return env.call(op, Bind(field), value)


def _sa_compare(op: str, column: sa.Column, value: Any):
    if op == 'eq':
        return column == value

    if op == 'ne':
        return column != value

    if op == 'lt':
        return column < value

    if op == 'le':
        return column <= value

    if op == 'gt':
        return column > value

    if op == 'ge':
        return column >= value

    if op == 'contains':
        return column.contains(value)

    if op == 'startswith':
        return column.startswith(value)

    raise NotImplementedError


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


T = TypeVar('T')


def _prepare_value(prop: Property, value: T) -> Union[T, List[T]]:
    return prepare_enum_value(prop, value)


@ufunc.resolver(SqlQueryBuilder, Bind, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, field: Bind, value: Any):
    prop = env.resolve_property(field)
    value = _prepare_value(prop, value)
    return env.call(op, prop.dtype, value)


@ufunc.resolver(SqlQueryBuilder, GetAttr, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, attr: GetAttr, value: Any):
    resolved = env.resolve_property(attr)
    return env.call(op, resolved, value)


@ufunc.resolver(SqlQueryBuilder, Bind, list, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, field: Bind, value: List[Any]):
    prop = env.resolve_property(field)
    value = list(flatten(_prepare_value(prop, v) for v in value))
    return env.call(op, prop.dtype, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, fpr: ForeignProperty, value: Any):
    value = _prepare_value(fpr.right.prop, value)
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, list, names=COMPARE)
def compare(
    env: SqlQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    value: List[Any],
):
    value = list(flatten(_prepare_value(fpr.right.prop, v) for v in value))
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(SqlQueryBuilder, DataType, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, dtype: DataType, value: Any):
    column = env.backend.get_column(env.table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, UUID, str, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, dtype: DataType, value: Any):
    column = env.backend.get_column(env.table, dtype.prop).cast(sa.String)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, Text, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, dtype: Text, value: Any):
    column = get_language_column(env, dtype)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, object, names=COMPARE)
def compare(
    env: SqlQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DataType,
    value: Any,
):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(SqlQueryBuilder, Function, object, names=COMPARE)
def compare(env: SqlQueryBuilder, op: str, func: Function, value: Any):
    return _sa_compare(op, func, value)


@ufunc.resolver(SqlQueryBuilder, DataType, list)
def eq(env: SqlQueryBuilder, dtype: DataType, value: List[Any]):
    column = env.backend.get_column(env.table, dtype.prop)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, Text, list)
def eq(env: SqlQueryBuilder, dtype: Text, value: List[Any]):
    column = get_language_column(env, dtype)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, list)
def eq(env: SqlQueryBuilder, fpr: ForeignProperty, dtype: DataType, value: list):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop)
    return column.in_(value)


@ufunc.resolver(SqlQueryBuilder, DataType, list)
def ne(env: SqlQueryBuilder, dtype: DataType, value: List[Any]):
    column = env.backend.get_column(env.table, dtype.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, Text, list)
def ne(env: SqlQueryBuilder, dtype: Text, value: List[Any]):
    column = get_language_column(env, dtype)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType, list)
def ne(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: List[Any],
):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, fpr.right.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, list)
def ne(env: SqlQueryBuilder, fpr: ForeignProperty, value: List[Any]):
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, fpr.right.prop)
    return ~column.in_(value)


@ufunc.resolver(SqlQueryBuilder, object)
def _resolve_unresolved(env: SqlQueryBuilder, value: Any) -> Any:
    if isinstance(value, Unresolved):
        raise ValueError(f"Unresolved value {value!r}.")
    else:
        return value


@ufunc.resolver(SqlQueryBuilder, Property)
def _resolve_unresolved(env: SqlQueryBuilder, prop: Property) -> sa.Column:
    if prop.external is None:
        raise NoExternalName(prop)

    if prop.external.name:
        return env.backend.get_column(env.table, prop)

    if prop.external.prepare and isinstance(prop.external.prepare, Expr):
        result = env(this=prop).resolve(prop.external.prepare)
        columns = env.call('_resolve_unresolved', result)
        return columns

    raise NoExternalName(prop)


@ufunc.resolver(SqlQueryBuilder, Bind)
def _resolve_unresolved(env: SqlQueryBuilder, field: Bind) -> sa.Column:
    prop = env.resolve_property(field)
    return env.call('_resolve_unresolved', prop)


@ufunc.resolver(SqlQueryBuilder, GetAttr)
def _resolve_unresolved(env: SqlQueryBuilder, attr: GetAttr) -> sa.Column:
    fpr = env.resolve_property(attr)
    table = env.joins.get_table(env, fpr)
    dtype = fpr.right
    return env.backend.get_column(table, dtype.prop)


@ufunc.resolver(SqlQueryBuilder, list)
def _resolve_unresolved(env: SqlQueryBuilder, data: list) -> list:
    return list(env.call('_resolve_unresolved', value) for value in data)


@ufunc.resolver(SqlQueryBuilder, tuple)
def _resolve_unresolved(env: SqlQueryBuilder, data: tuple) -> tuple:
    return tuple(env.call('_resolve_unresolved', value) for value in data)


@ufunc.resolver(SqlQueryBuilder, Expr, name='and')
def and_(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    args = [
        env.call('_resolve_unresolved', arg)
        for arg in args
        if arg is not None
    ]
    if len(args) > 1:
        return sa.and_(*args)
    elif args:
        return args[0]


@ufunc.resolver(SqlQueryBuilder, Expr, name='or')
def or_(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    args = [
        env.call('_resolve_unresolved', arg)
        for arg in args
        if arg is not None
    ]
    if len(args) > 1:
        return sa.or_(*args)
    elif args:
        return args[0]


@ufunc.resolver(SqlQueryBuilder, Expr, name='list')
def list_(env: SqlQueryBuilder, expr: Expr) -> List[Any]:
    args, kwargs = expr.resolve(env)
    return list(args)


@ufunc.resolver(SqlQueryBuilder)
def count(env: SqlQueryBuilder):
    return sa.func.count()


def _get_property_for_select(
    env: SqlQueryBuilder,
    name: str,
    *,
    nested: bool = False,
) -> Property:
    # TODO: `name` can refer to (in specified order):
    #       - var - a defined variable
    #       - param - a parameter if parametrization is used
    #       - item - an item of a dict or list
    #       - prop - a property
    #       Currently only `prop` is resolved.
    prop = env.resolve_property(name)
    if (
        # Check authorization only for top level properties in select list.
        # XXX: Not sure if nested is the right property to user, probably better
        #      option is to check if this call comes from a prepare context. But
        #      then how prepare context should be defined? Probably resolvers
        #      should be called with a different env class?
        #      tag:resolving_private_properties_in_prepare_context
        nested or
        authorized(env.context, prop, Action.SEARCH)
    ):
        return prop
    else:
        raise PropertyNotFound(env.model, property=name)


@ufunc.resolver(SqlQueryBuilder, Expr)
def select(env: SqlQueryBuilder, expr: Expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())
    if env.selected is not None:
        raise RuntimeError("`select` was already called.")
    env.selected = {}
    if args:
        for key, arg in args:
            selected = env.call('select', arg)
            if selected is not None:
                env.selected[key] = selected
    else:
        for prop in take(['_id', all], env.model.properties).values():
            if authorized(env.context, prop, Action.GETALL):
                processed = env.call('select', prop)
                if not prop.dtype.inherited or processed.prep is not None:
                    env.selected[prop.place] = processed

    if not (len(args) == 1 and args[0][0] == '_page'):
        if not env.columns:
            raise RuntimeError(
                f"{expr} didn't added anything to select list."
            )


@ufunc.resolver(SqlQueryBuilder, sa.sql.expression.ColumnElement)
def select(env, column):
    return Selected(env.add_column(column))


@ufunc.resolver(SqlQueryBuilder, Bind)
def select(env: SqlQueryBuilder, item: Bind, *, nested: bool = False):
    if item.name == '_page':
        return None
    prop = _get_property_for_select(env, item.name, nested=nested)
    return env.call('select', prop)


@ufunc.resolver(SqlQueryBuilder, str)
def select(env: SqlQueryBuilder, item: str, *, nested: bool = False):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    prop = _get_property_for_select(env, item, nested=nested)
    return env.call('select', prop)


@ufunc.resolver(SqlQueryBuilder, Property)
def select(env: SqlQueryBuilder, prop: Property) -> Selected:
    if prop.place not in env.resolved:
        if isinstance(prop.external, list):
            raise SourceCannotBeList(prop)
        if prop.external.prepare is not NA:
            # If `prepare` formula is given, evaluate formula.
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = process_literal_value(prop.external.prepare)

            # XXX: Maybe interpretation of prepare formula should be done under
            #      a different Env? This way, select resolvers would know when
            #      properties are resolved under a formula context and for
            #      example would be able to decide to not check permissions on
            #      properties.
            #      tag:resolving_private_properties_in_prepare_context
            result = env.call('select', prop.dtype, result)
        elif prop.external and prop.external.name:
            # If prepare is not given, then take value from `source`.
            result = env.call('select', prop.dtype)
        elif prop.is_reserved():
            # Reserved properties never have external source.
            result = env.call('select', prop.dtype)
        elif not prop.dtype.requires_source:
            # Some DataTypes might have children that have source instead of themselves, like: Text, Object
            result = env.call('select', prop.dtype)
        elif prop.dtype.inherited:
            # Some DataTypes might be inherited, or hidden, so we need to go through them in case they can be joined
            result = env.call('select', prop.dtype)
            if not isinstance(result, Selected):
                result = Selected(prop=prop, prep=None)
        else:
            # If `source` is not given, return None.
            result = Selected(prop=prop, prep=None)
        assert isinstance(result, Selected), prop
        env.resolved[prop.place] = result
    return env.resolved[prop.place]


@ufunc.resolver(SqlQueryBuilder, DataType)
def select(env: SqlQueryBuilder, dtype: DataType) -> Selected:
    table = env.backend.get_table(env.model)
    column = env.backend.get_column(table, dtype.prop, select=True)
    return Selected(
        item=env.add_column(column),
        prop=dtype.prop,
    )


@ufunc.resolver(SqlQueryBuilder, Object)
def select(env: SqlQueryBuilder, dtype: Object) -> Selected:
    prep = {}
    for prop in take(dtype.properties).values():
        sel = env.call('select', prop)
        if sel is not None:
            prep[prop.name] = sel
    return Selected(prop=dtype.prop, prep=prep)


@ufunc.resolver(SqlQueryBuilder, Ref)
def select(env: SqlQueryBuilder, dtype: Ref) -> Selected:
    prep = {}
    if not dtype.inherited:
        prep['_id'] = Selected(prop=dtype.prop, prep=select_ref_foreign_key_properties(env, dtype))

    for prop in dtype.properties.values():
        sel = env.call('select', prop)
        prep[prop.name] = sel

    return Selected(
        prop=dtype.prop,
        prep=prep
    )


@ufunc.resolver(SqlQueryBuilder, ExternalRef)
def select(env: SqlQueryBuilder, dtype: ExternalRef) -> Selected:
    prep = {}
    if not dtype.inherited:
        prep.update(select_external_ref_foreign_key_properties(env, dtype))

    for prop in dtype.properties.values():
        sel = env.call('select', prop)
        prep[prop.name] = sel

    return Selected(
        prop=dtype.prop,
        prep=prep
    )


@ufunc.resolver(SqlQueryBuilder, Ref, ReservedProperty)
def select(env: SqlQueryBuilder, dtype: Ref, reserved: ReservedProperty):
    return Selected(prop=dtype.prop, prep=select_ref_foreign_key_properties(env, dtype))


@ufunc.resolver(SqlQueryBuilder, ExternalRef, ReservedProperty)
def select(env: SqlQueryBuilder, dtype: ExternalRef, reserved: ReservedProperty):
    prep = select_external_ref_foreign_key_properties(env, dtype, target=reserved.param)
    return prep[reserved.param]


@ufunc.resolver(SqlQueryBuilder, Ref, str)
def select(env, dtype: Ref, prop: str):
    table = env.backend.get_table(env.model)
    return Selected(
        env.add_column(
            table.c[dtype.prop.place + '.' + prop]
        )
    )


@ufunc.resolver(SqlQueryBuilder, Ref, (list, tuple))
def select(env: SqlQueryBuilder, dtype: Ref, data: list | tuple) -> Selected:
    prep = {}
    if not dtype.inherited:
        prep['_id'] = Selected(
            prop=dtype.prop,
            prep=select_ref_foreign_key_properties(env, dtype, properties=data)
        )

    for prop in dtype.properties.values():
        sel = env.call('select', prop)
        prep[prop.name] = sel

    return Selected(
        prop=dtype.prop,
        prep=prep
    )


@ufunc.resolver(SqlQueryBuilder, ExternalRef, (list, tuple))
def select(env: SqlQueryBuilder, dtype: ExternalRef, data: list | tuple) -> Selected:
    prep = {}
    if not dtype.inherited:
        prep.update(select_external_ref_foreign_key_properties(env, dtype, properties=data))

    for prop in dtype.properties.values():
        sel = env.call('select', prop)
        prep[prop.name] = sel

    return Selected(
        prop=dtype.prop,
        prep=prep
    )


@ufunc.resolver(SqlQueryBuilder, String)
def select(env, dtype):
    env.call('validate_dtype_for_select', dtype, _gather_selected_properties(env))
    column = env.backend.get_column(env.table, dtype.prop)
    return Selected(env.add_column(column), dtype.prop)


@ufunc.resolver(SqlQueryBuilder, Text)
def select(env, dtype):
    env.call('validate_dtype_for_select', dtype, _gather_selected_properties(env))
    if env.query_params.push:
        result = {
            key: env.call('select', prop)
            for key, prop in dtype.langs.items()
        }
        return Selected(prop=dtype.prop, prep=result)

    if env.query_params.lang:
        for lang in env.query_params.lang:
            if isinstance(lang, Star):
                result = {
                    key: env.call('select', prop)
                    for key, prop in dtype.langs.items()
                }
                return Selected(prop=dtype.prop, prep=result)
            break

        result = {
            key: env.call('select', prop)
            for key, prop in dtype.langs.items() if key in env.query_params.lang
        }
        return Selected(prop=dtype.prop, prep=result)

    default_langs = env.context.get('config').languages
    lang_prop = determine_language_property_for_text(dtype, env.query_params.lang_priority, default_langs)
    return Selected(prop=dtype.prop, prep={
        lang_prop.name: env.call('select', lang_prop)
    })


@ufunc.resolver(SqlQueryBuilder, Denorm)
def select(env, dtype: Denorm):
    fpr = env.call('_denorm_to_foreign_property', dtype)
    return env.call("select", fpr)


@ufunc.resolver(SqlQueryBuilder, Array)
def select(env: SqlQueryBuilder, dtype: Array):
    if dtype.model is not None:
        table = env.joins.get_intermediate_table(env, dtype)
        right = dtype.right_prop
        if right.external is None:
            raise NoExternalName(right)

        columns = env(table=table, model=dtype.model).call('_resolve_unresolved', right)
        column = env.call('_group_array', columns)

        # Group by all (required for aggregation functions)
        env.add_to_group_by(list(env.table.columns))
        return Selected(
            prop=dtype.prop,
            prep=Selected(
                item=env.add_column(column),
                prop=dtype.right_prop
            )
        )

    if dtype.prop.external.name and dtype.prop.external.prepare:
        table = env.backend.get_table(dtype.prop.model)
        column = env.backend.get_column(table, dtype.prop)
        return Selected(
            item=env.add_column(column),
            prop=dtype.prop,
            prep=env.call('select', dtype.items)
        )

    return Selected(
        prop=dtype.prop,
        prep=env.call('select', dtype.items)
    )


@ufunc.resolver(SqlQueryBuilder, Page)
def select(env: SqlQueryBuilder, page: Page) -> List[sa.Column]:
    table = env.backend.get_table(env.model)
    return_selected = []
    for item in page.by.values():
        column = env.backend.get_column(table, item.prop, select=True)
        return_selected.append(column)
    return return_selected


@ufunc.resolver(SqlQueryBuilder, PrimaryKey)
def select(
    env: SqlQueryBuilder,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys
    if not pkeys:
        # If primary key is not specified use all properties to uniquely
        # identify row.
        pkeys = take(model.properties).values()

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', prop)
    else:
        result = [
            env.call('select', prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, DataType)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
) -> Selected:
    table = env.joins.get_table(env, fpr)
    column = env.backend.get_column(table, dtype.prop, select=True)
    return Selected(
        item=env.add_column(column),
        prop=dtype.prop,
    )


@ufunc.resolver(SqlQueryBuilder, Property)
def join_table_on(env: SqlQueryBuilder, prop: Property) -> Any:
    if prop.external.prepare is not NA:
        if isinstance(prop.external.prepare, Expr):
            result = env.resolve(prop.external.prepare)
        else:
            result = process_literal_value(prop.external.prepare)
        return env.call('join_table_on', prop.dtype, result)
    else:
        return env.call('join_table_on', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType)
def join_table_on(env: SqlQueryBuilder, dtype: DataType) -> Tuple[Any]:
    table = env.table
    if table is None:
        table = env.backend.get_table(dtype.prop.model)
    column = env.backend.get_column(table, dtype.prop)
    return column


@ufunc.resolver(SqlQueryBuilder, Text)
def join_table_on(env: SqlQueryBuilder, dtype: Text) -> Tuple[Any]:
    column = get_language_column(env, dtype)
    return column


@ufunc.resolver(SqlQueryBuilder, PrimaryKey)
def join_table_on(env: SqlQueryBuilder, dtype: DataType) -> Any:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('join_table_on', prop)
    else:
        result = [
            env.call('join_table_on', prop)
            for prop in pkeys
        ]

    return result


@ufunc.resolver(SqlQueryBuilder, Ref)
def join_table_on(
    env: SqlQueryBuilder,
    dtype: Ref,
) -> Tuple[Any]:
    original_model = dtype.prop.model
    original_table = env.backend.get_table(original_model)

    prop_name = dtype.prop.name
    parent_ref = None
    if dtype.inherited:
        parent: Property = dtype.prop.parent
        parent_ref = parent.dtype.model.properties.get(prop_name)
    filtered_properties = {prop.name: prop for prop in dtype.properties.values() if not prop.dtype.inherited}
    ref_properties = {}
    # Gather overwritten properties
    # missing ref properties are None
    for prop in dtype.refprops:
        prop_name = prop.name
        value = filtered_properties.get(prop_name, None)
        if value is not None:
            value = env(table=original_table, model=original_model).call('join_table_on', value)
        ref_properties[prop_name] = value

    # Fill missing None values using parent reference (from which it was inherited from)
    if any(value is None for value in ref_properties.values()):
        if parent_ref is None:
            table = env.table
            column = env.backend.get_column(table, dtype.prop)
            return column

        parent_joined_columns = env(model=parent_ref.model).call('join_table_on', parent_ref)
        for i, (key, value) in enumerate(ref_properties.items()):
            if value is None:
                ref_properties[key] = parent_joined_columns[i]
    return tuple(value for value in ref_properties.values())


@ufunc.resolver(SqlQueryBuilder, DataType, tuple)
def join_table_on(
    env: SqlQueryBuilder,
    dtype: DataType,
    prep: tuple,
) -> Tuple[Any]:
    return tuple(env.call('join_table_on', v) for v in prep)


@ufunc.resolver(SqlQueryBuilder, Bind)
def join_table_on(env: SqlQueryBuilder, item: Bind):
    prop = env.resolve_property(item)
    if not authorized(env.context, prop, Action.SEARCH):
        raise PropertyNotFound(env.model, property=item.name)
    return env.call('join_table_on', prop)


@ufunc.resolver(SqlQueryBuilder, LiteralProperty)
def join_table_on(env: SqlQueryBuilder, item: LiteralProperty):
    return item.value


@ufunc.resolver(SqlQueryBuilder, DataType, LiteralProperty)
def join_table_on(env: SqlQueryBuilder, dtype: DataType, item: LiteralProperty):
    return env.call('join_table_on', item)


@ufunc.resolver(SqlQueryBuilder, Bind, name='len')
def len_(env: SqlQueryBuilder, bind: Bind):
    prop = env.resolve_property(bind)
    return env.call('len', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, str, name='len')
def len_(env: SqlQueryBuilder, bind: str):
    # XXX: Backwards compatible resolver, `str` arguments are deprecated.
    prop = env.resolve_property(bind)
    return env.call('len', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType, name='len')
def len_(env: SqlQueryBuilder, dtype: DataType):
    column = env.backend.get_column(env.table, dtype.prop)
    return sa.func.length(column)


@ufunc.resolver(SqlQueryBuilder, Expr)
def sort(env: SqlQueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    env.sort = []
    for key in args:
        result = env.call('sort', key)
        if isinstance(result, (list, set, tuple)):
            env.sort += result
        else:
            env.sort.append(result)


@ufunc.resolver(SqlQueryBuilder, DataType)
def sort(env, dtype):
    return env.call('asc', dtype)


@ufunc.resolver(SqlQueryBuilder, Bind)
def sort(env, field):
    prop = env.resolve_property(field)
    return env.call('asc', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, Negative)
def sort(env, field):
    prop = env.resolve_property(field)
    return env.call('desc', prop.dtype)


@ufunc.resolver(SqlQueryBuilder, DataType)
def asc(env, dtype):
    column = env.backend.get_column(env.table, dtype.prop)
    return env.call('asc', column)


@ufunc.resolver(SqlQueryBuilder, Text)
def asc(env, dtype):
    column = get_language_column(env, dtype)
    return env.call('asc', column)


# Reason for column == None is NULLS LAST, because
# if it's NULL it will return 1 and if it's not NULL it will return 0
# when ordering 0 takes priority over 1, so then it will sort first values that are not NULL
@ufunc.resolver(SqlQueryBuilder, sa.sql.expression.ColumnElement)
def asc(env, column):
    return sa.case(
        [
            (
                column == None,
                sa.literal_column('1', type_=sa.Integer)
            )
        ],
        else_=sa.literal_column('0', type_=sa.Integer)
    ), column.asc()


@ufunc.resolver(SqlQueryBuilder, DataType)
def desc(env, dtype):
    column = env.backend.get_column(env.table, dtype.prop)
    return env.call('desc', column)


@ufunc.resolver(SqlQueryBuilder, Text)
def desc(env, dtype):
    column = get_language_column(env, dtype)
    return env.call('desc', column)


# Reason for column != None is NULLS FIRST, because
# if it's NULL it will return 0 and if it's not NULL it will return 1
# when ordering 0 takes priority over 1, so then it will sort first values that are NULL
@ufunc.resolver(SqlQueryBuilder, sa.sql.expression.ColumnElement)
def desc(env, column):
    return sa.case(
        [
            (
                column != None,
                sa.literal_column('1', type_=sa.Integer)
            )
        ],
        else_=sa.literal_column('0', type_=sa.Integer)
    ), column.desc()


@ufunc.resolver(SqlQueryBuilder, DataType)
def negative(env: SqlQueryBuilder, dtype: DataType):
    return Negative(dtype.prop.place)


@ufunc.resolver(SqlQueryBuilder, GetAttr)
def negative(env: SqlQueryBuilder, attr: GetAttr):
    resolved = env.resolve_property(attr)
    return env.call('negative', resolved)


@ufunc.resolver(SqlQueryBuilder, String)
def negative(env: SqlQueryBuilder, dtype: String):
    if dtype.prop.parent and isinstance(dtype.prop.parent.dtype, Text):
        return Negative(dtype.prop.place.replace('.', '@'))
    return Negative(dtype.prop.place)


@ufunc.resolver(SqlQueryBuilder, int)
def limit(env: SqlQueryBuilder, n: int):
    env.limit = n


@ufunc.resolver(SqlQueryBuilder, int)
def offset(env: SqlQueryBuilder, n: int):
    env.offset = n


@ufunc.resolver(SqlQueryBuilder, Expr)
def file(env: SqlQueryBuilder, expr: Expr) -> Expr:
    args, kwargs = expr.resolve(env)
    return Expr(
        'file',
        name=env.call('select', kwargs['name'], nested=True),
        content=env.call('select', kwargs['content'], nested=True),
    )


@overload
@ufunc.resolver(SqlQueryBuilder)
def cast(env: SqlQueryBuilder) -> Expr:
    return Expr('cast')


@overload
@ufunc.resolver(SqlQueryBuilder, Bind, Bind)
def point(env: SqlQueryBuilder, x: Bind, y: Bind) -> Expr:
    return Expr(
        'point',
        env.call('select', x, nested=True),
        env.call('select', y, nested=True),
    )


@ufunc.resolver(SqlQueryBuilder)
def distinct(env: SqlQueryBuilder):
    if env.model and env.model.external and not env.model.external.unknown_primary_key:
        extracted_columns = [env.backend.get_column(env.table, prop) for prop in env.model.external.pkeys if
                             prop.external]
        env.add_to_group_by(extracted_columns)
    else:
        env.distinct = True


@ufunc.resolver(SqlQueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: SqlQueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    super_ = ufunc.resolver[env, fpr, dtype]
    return super_(env, fpr, dtype)


@ufunc.resolver(SqlQueryBuilder, object)
def _group_array(env: SqlQueryBuilder, columns: Any):
    raise NotImplementedFeature(env.backend, feature="Ability to group arrays using default `sql` type")



import contextlib

import sqlalchemy as sa
from sqlalchemy.engine.base import Engine

from spinta import commands
from spinta.backends.constants import BackendFeatures
from spinta.components import Model
from spinta.components import Property
from spinta.datasets.components import ExternalBackend
from spinta.exceptions import BackendUnavailable


class Sql(ExternalBackend):
    type: str = 'sql'
    engine: Engine = None
    schema: sa.MetaData = None
    dbschema: str = None  # Database schema name

    features = {
        BackendFeatures.PAGINATION
    }

    query_builder_type = 'sql'
    result_builder_type = 'sql'

    @contextlib.contextmanager
    def transaction(self, write=False):
        raise NotImplementedError

    @contextlib.contextmanager
    def begin(self):
        try:
            with self.engine.begin() as conn:
                yield conn
        except sa.exc.OperationalError:
            self.available = False
            raise BackendUnavailable(self)

    def get_table(self, model: Model, name: str = None) -> sa.Table:
        name = name or model.external.name

        if self.dbschema:
            key = f'{self.dbschema}.{name}'
        else:
            key = name

        if key not in self.schema.tables:
            sa.Table(name, self.schema, autoload_with=self.engine)

        return self.schema.tables[key]

    def get_column(
        self,
        table: sa.Table,
        prop: Property,
        *,
        select=False,
        **kwargs
    ) -> sa.Column:
        column = commands.get_column(self, prop, table=table, **kwargs)
        return column



from typing import List, Dict
from typer import echo

from spinta.cli.helpers.upgrade.clients import migrate_clients, cli_requires_clients_migration
from spinta.cli.helpers.upgrade.components import UpgradeComponent, UPGRADE_CLIENTS_SCRIPT
from spinta.cli.helpers.upgrade.helpers import script_check_status_message
from spinta.components import Context

# Global upgrade script mapper
# To add new script, just create script name and create `UpgradeComponent` for it
# If there are no checks needed, just pass None to `check`
__upgrade_scripts: Dict[str, UpgradeComponent] = {
    UPGRADE_CLIENTS_SCRIPT: UpgradeComponent(
        upgrade=migrate_clients,
        check=cli_requires_clients_migration
    )
}

UPGRADE_CHECK_STATUS_PASSED = "PASSED"
UPGRADE_CHECK_STATUS_REQUIRED = "REQUIRED"
UPGRADE_CHECK_STATUS_FORCED = "FORCED"


# Returns a list of all available script names
def get_all_upgrade_script_names() -> List[str]:
    return list(__upgrade_scripts.keys())


def run_all_scripts(
    context: Context,
    destructive: bool = False,
    force: bool = False,
    **kwargs
):
    for script_name in __upgrade_scripts.keys():
        run_specific_script(
            context=context,
            script_name=script_name,
            destructive=destructive,
            force=force,
            **kwargs
        )


def does_script_exist(
    script_name: str
) -> bool:
    return script_name in __upgrade_scripts


def run_specific_script(
    context: Context,
    script_name: str,
    destructive: bool = False,
    force: bool = False,
    **kwargs
):
    script = __upgrade_scripts[script_name]

    status = UPGRADE_CHECK_STATUS_PASSED
    if force or script.check(context, **kwargs):
        status = UPGRADE_CHECK_STATUS_FORCED if force else UPGRADE_CHECK_STATUS_REQUIRED
        echo(script_check_status_message(script_name, status))
        script.upgrade(context, destructive=destructive, **kwargs)
    else:
        echo(script_check_status_message(script_name, status))



from __future__ import annotations

from spinta.datasets.backends.sql.ufuncs.query.components import SqlQueryBuilder
from spinta.datasets.backends.sql.backends.oracle.components import Oracle


class OracleQueryBuilder(SqlQueryBuilder):
    backend: Oracle




import os
import pathlib
import uuid
from typing import Any
from typer import echo
import ruamel.yaml

from spinta.cli.helpers.upgrade.components import UPGRADE_CLIENTS_SCRIPT
from spinta.cli.helpers.upgrade.helpers import script_destructive_warning
from spinta.components import Context, Config
from spinta.core.config import DEFAULT_CONFIG_PATH
from spinta.utils.config import get_clients_path, get_keymap_path, get_id_path, get_helpers_path
from spinta.utils.types import is_str_uuid

from authlib.oauth2.rfc6749.errors import InvalidClientError

yaml = ruamel.yaml.YAML(typ="safe")

yml = ruamel.yaml.YAML()
yml.indent(mapping=2, sequence=4, offset=2)
yml.width = 80
yml.explicit_start = False

CLIENT_STATUS_SUCCESS = "SUCCESS"
CLIENT_STATUS_SKIPPED_MIGRATED = "SKIPPED (ALREADY MIGRATED)"
CLIENT_STATUS_FAILED_INVALID = "FAILED (INVALID STRUCTURE)"
CLIENT_STATUS_FAILED_MISSING_ID = "FAILED (MISSING `client_id` FIELD)"
CLIENT_STATUS_FAILED_ID_NOT_UUID = "FAILED (`client_id` MUST BE UUID)"
CLIENT_STATUS_FAILED_MISSING_SECRET = "FAILED (MISSING `client_secret_hash` FIELD)"
CLIENT_STATUS_FAILED_MISSING_SCOPES = "FAILED (MISSING `scopes` FIELD)"
CLIENT_STATUS_FAILED_MISSING_NAME = "FAILED (MISSING `client_name` FIELD)"


def migrate_clients(context: Context, destructive: bool, **kwargs: Any):
    config: Config = context.get('config')
    config_path = config.config_path or DEFAULT_CONFIG_PATH
    config_path = pathlib.Path(config_path)

    clients_path = get_clients_path(config_path)

    helpers_path = get_helpers_path(clients_path)
    keymap_path = get_keymap_path(clients_path)
    id_path = get_id_path(clients_path)

    # Ensure clients path exists
    os.makedirs(clients_path, exist_ok=True)

    # Ensure helpers path exists
    os.makedirs(helpers_path, exist_ok=True)

    # Ensure keymap.yml exists
    keymap_path.touch(exist_ok=True)

    # Ensure id path exists
    os.makedirs(id_path, exist_ok=True)

    keymap = yaml.load(keymap_path)
    if keymap is None:
        keymap = {}

    if destructive:
        echo(script_destructive_warning(UPGRADE_CLIENTS_SCRIPT, "override already migrated files with old ones"))

    items = os.listdir(clients_path)
    for item in items:
        if item.endswith('.yml'):
            status = _migrate_client_file(
                client_file_path=clients_path / item,
                id_path=id_path,
                destructive=destructive,
                keymap=keymap
            )
            echo(client_migration_status_message(item, status))

    _recreate_keymap(
        id_path=id_path,
        keymap_path=keymap_path
    )


def _recreate_keymap(
    id_path: pathlib.Path,
    keymap_path: pathlib.Path
):
    id_items = os.listdir(id_path)
    keymap = {}
    for id0 in id_items:
        if len(id0) != 2:
            continue

        id0_items = os.listdir(id_path / id0)
        for id1 in id0_items:
            if len(id1) != 2:
                continue

            id1_items = os.listdir(id_path / id0 / id1)
            for uuid_item in id1_items:
                if uuid_item.endswith('.yml') and len(uuid_item) == 36:
                    try:
                        data = yaml.load(id_path / id0 / id1 / uuid_item)
                        keymap[data["client_name"]] = data["client_id"]
                    except FileNotFoundError:
                        raise (InvalidClientError(description='Invalid client id or secret'))

    echo(f"Created keymap with {len(keymap)} users")
    yml.dump(keymap, keymap_path)


def _generate_new_file_path(
    id_path: pathlib.Path,
    client_id: str,
    with_yml: bool = True
) -> pathlib.Path:
    folder_path = id_path / client_id[:2] / client_id[2:4]
    if with_yml:
        return folder_path / f'{client_id[4:]}.yml'
    return folder_path


def _validate_file_structure(
    data: dict,
    old: bool = True
) -> (bool, str):
    if not isinstance(data, dict) or data == {}:
        return False, CLIENT_STATUS_FAILED_INVALID

    # All versions must contain 'client_id' field
    if 'client_id' not in data:
        return False, CLIENT_STATUS_FAILED_MISSING_ID
    else:
        # If it's not old, then 'client_id' must be uuid
        if not old and is_str_uuid(data["client_id"]):
            return False, CLIENT_STATUS_FAILED_ID_NOT_UUID

    # All versions must contain 'client_secret_hash'
    if 'client_secret_hash' not in data:
        return False, CLIENT_STATUS_FAILED_MISSING_SECRET

    # All versions must contain 'scopes
    if 'scopes' not in data:
        return False, CLIENT_STATUS_FAILED_MISSING_SCOPES

    # Only new version must contain 'client_name'
    if not old and 'client_name' not in data:
        return False, CLIENT_STATUS_FAILED_MISSING_NAME

    return True, ""


def _migrate_client_file(
    client_file_path: pathlib.Path,
    id_path: pathlib.Path,
    destructive: bool,
    keymap: dict
) -> str:
    try:
        old_data = yaml.load(client_file_path)
    except FileNotFoundError:
        raise (InvalidClientError(description='Could not open client file'))

    validated = _validate_file_structure(old_data)
    if not validated[0]:
        return validated[1]

    client_name = old_data.get("client_name", old_data["client_id"])
    client_id = old_data['client_id']

    # Check if new valid client file exists, if so, skip it
    if client_name in keymap:
        keymap_value = keymap[client_name]
        new_file_path = _generate_new_file_path(
            id_path,
            keymap_value
        )
        client_id = keymap_value
        if new_file_path.exists() and not destructive:
            return CLIENT_STATUS_SKIPPED_MIGRATED

    if not is_str_uuid(client_id):
        client_id = str(uuid.uuid4())

    new_path = _generate_new_file_path(id_path, client_id)
    os.makedirs(_generate_new_file_path(id_path, client_id, False), exist_ok=True)

    data = {
        "client_id": client_id,
        "client_name": client_name,
        "client_secret_hash": old_data["client_secret_hash"],
        "scopes": old_data["scopes"]
    }
    yml.dump(data, new_path)
    return CLIENT_STATUS_SUCCESS


def cli_requires_clients_migration(context: Context, **kwargs: Any) -> bool:
    config: Config = context.get('config')
    config_path = config.config_path or DEFAULT_CONFIG_PATH
    config_path = pathlib.Path(config_path)

    clients_path = get_clients_path(config_path)
    return requires_client_migration(clients_path)


def requires_client_migration(clients_path: pathlib.Path) -> bool:
    # Cannot apply any migrations, if there are no files
    if not clients_path.exists():
        return False

    items = os.listdir(clients_path)
    contains_yml = any([item.endswith('.yml') for item in items])
    if contains_yml:
        keymap_path = get_keymap_path(clients_path)

        # If there are yml files, there should /helpers/keymap.yml
        if not keymap_path.exists():
            return True

        # If there are yml files, there should be /id folder
        id_path = get_id_path(clients_path)
        if not id_path.exists():
            return True

        keymap = yaml.load(keymap_path)
        # Keymap should not be empty or None
        if not keymap:
            return True

        ids = os.listdir(id_path)
        # id folder should contain subfolders that are 2 characters long (meaning there are migrated files)
        if not any(len(item) == 2 for item in ids):
            return True

    return False


def client_migration_status_message(file_name: str, status: str) -> str:
    return f"\tMigrating {file_name.ljust(40)}\tStatus: {status}"



from typing import List, Union

import sqlalchemy as sa
from sqlalchemy.sql.type_api import TypeEngine

from spinta import commands
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name
from spinta.backends.postgresql.helpers.name import get_pg_column_name, get_pg_table_name
from spinta.components import Context, Property
from spinta.types.datatype import Ref


@commands.prepare.register(Context, PostgreSQL, Ref)
def prepare(context: Context, backend: PostgreSQL, dtype: Ref, propagate: bool = True, **kwargs):
    pkey_type = commands.get_primary_key_type(context, backend)
    columns = []

    if not dtype.inherited:
        columns = get_pg_foreign_key(
            dtype.prop,
            table_name=get_pg_table_name(get_table_name(dtype.model)),
            model_name=dtype.prop.model.name,
            column_type=pkey_type,
        )

    if not propagate:
        return columns

    for key, value in dtype.properties.items():
        res = commands.prepare(context, backend, value, **kwargs)
        if res is not None:
            if isinstance(res, list):
                columns.extend(res)
            else:
                columns.append(res)
    return columns


def get_pg_foreign_key(
    prop: Property,
    *,
    table_name: str,
    model_name: str,
    column_type: TypeEngine
) -> List[Union[sa.Column, sa.Constraint]]:
    column_name = get_pg_column_name(get_column_name(prop) + '._id')
    if prop.list and prop.place == prop.list.place:
        column_name = '_id'
    nullable = not prop.dtype.required

    columns = [
        sa.Column(
            column_name,
            column_type,
            nullable=nullable,
            index=not prop.dtype.unique,
            unique=prop.dtype.unique
        ),
        sa.ForeignKeyConstraint(
            [column_name], [f'{table_name}._id'],
            ondelete='CASCADE' if prop.list else None,
            onupdate='CASCADE' if prop.list else None
        )
    ]
    return columns



from enum import Enum

DATASET = [
    'id',
    'dataset',
    'resource',
    'base',
    'model',
    'property',
    'type',
    'ref',
    'source',
    'prepare',
    'level',
    'access',
    'uri',
    'title',
    'description',
]


class DataTypeEnum(Enum):
    STRING = 'string'
    TEXT = 'text'
    REF = 'ref'
    BACKREF = 'backref'
    ARRAY = 'array'
    GENERIC = 'generic'
    INHERIT = 'inherit'
    _OBJECT = 'object'                # Internal type
    _PARTIAL = 'partial'              # Internal type
    _ARRAY_BACKREF = 'array_backref'  # Internal type
    _PARTIAL_ARRAY = 'partial_array'  # Internal type




from spinta.backends.postgresql.ufuncs.query.components import InheritForeignProperty
from spinta.components import Property, FuncProperty
from spinta.core.ufuncs import ufunc, Bind, GetAttr
from spinta.exceptions import PropertyNotFound, FieldNotInResource, LangNotDeclared
from spinta.types.datatype import DataType, Ref, BackRef, Object, Array, ExternalRef, File, Inherit, Denorm
from spinta.types.text.components import Text
from spinta.ufuncs.components import ForeignProperty
from spinta.ufuncs.propertyresolver.components import PropertyResolver
from spinta.ufuncs.querybuilder.components import ReservedProperty, NestedProperty


@ufunc.resolver(PropertyResolver, Property, object)
def _resolve_property(
    env: PropertyResolver,
    prop: Property,
    attr: object
):
    return env.call('_resolve_property', prop.dtype, attr)


@ufunc.resolver(PropertyResolver, DataType, object)
def _resolve_property(
    env: PropertyResolver,
    dtype: DataType,
    attr: object
):
    raise Exception("Cannot map", dtype.prop, attr)


@ufunc.resolver(PropertyResolver, Ref, GetAttr)
def _resolve_property(
    env: PropertyResolver,
    dtype: Ref,
    attr: GetAttr
):
    if attr.obj in dtype.properties:
        prop = dtype.properties[attr.obj]
    else:
        if attr.obj not in dtype.model.properties:
            raise FieldNotInResource(dtype, property=attr.obj)

        prop = dtype.model.properties[attr.obj]

    if env.ufunc_types:
        fpr = ForeignProperty(None, dtype, prop.dtype)
        return env.call('_resolve_property', fpr, prop, attr.name)
    return env.call('_resolve_property', prop, attr.name)


@ufunc.resolver(PropertyResolver, Ref, Bind)
def _resolve_property(
    env: PropertyResolver,
    dtype: Ref,
    attr: Bind
):
    if attr.name in dtype.properties:
        prop = dtype.properties[attr.name]
        return prop

    if attr.name not in dtype.model.properties:
        raise FieldNotInResource(dtype, property=attr.name)
    prop = dtype.model.properties[attr.name]

    if env.ufunc_types:
        # Check for self reference, no need to do joins if table already contains the value
        if attr.name == '_id':
            return ReservedProperty(dtype, attr.name)

        return ForeignProperty(None, dtype, prop.dtype)

    return prop


@ufunc.resolver(PropertyResolver, ExternalRef, Bind)
def _resolve_property(
    env: PropertyResolver,
    dtype: Ref,
    attr: Bind
):
    if attr.name in dtype.properties:
        prop = dtype.properties[attr.name]
        return prop

    if attr.name not in dtype.model.properties:
        raise FieldNotInResource(dtype, property=attr.name)
    prop = dtype.model.properties[attr.name]

    if env.ufunc_types:
        # Check for self reference, no need to do joins if table already contains the value
        for refprop in dtype.refprops:
            if refprop.name == attr.name:
                return ReservedProperty(dtype, attr.name)

        return ForeignProperty(None, dtype, prop.dtype)

    return prop


@ufunc.resolver(PropertyResolver, BackRef, GetAttr)
def _resolve_property(
    env: PropertyResolver,
    dtype: BackRef,
    attr: GetAttr,
):
    if attr.obj in dtype.properties:
        prop = dtype.properties[attr.obj]
    else:
        if attr.obj not in dtype.model.properties:
            raise FieldNotInResource(dtype, property=attr.obj)

        prop = dtype.model.properties[attr.obj]

    if env.ufunc_types:
        fpr = ForeignProperty(None, dtype, prop.dtype)
        return env.call('_resolve_property', fpr, prop.dtype, attr.name)

    return env.call('_resolve_property', prop, attr.name)


@ufunc.resolver(PropertyResolver, BackRef, Bind)
def _resolve_property(env, dtype, attr):
    if attr.obj in dtype.properties:
        return dtype.properties[attr.obj]

    if attr.obj not in dtype.model.properties:
        raise FieldNotInResource(dtype, property=attr.obj)

    prop = dtype.model.properties[attr.obj]
    if env.ufunc_types:
        fpr = ForeignProperty(None, dtype, prop.dtype)
        return env.call('_resolve_property', fpr, prop.dtype, attr.name)

    return prop


@ufunc.resolver(PropertyResolver, Object, GetAttr)
def _resolve_property(
    env: PropertyResolver,
    dtype: Object,
    attr: GetAttr,
):
    if attr.obj in dtype.properties:
        prop = dtype.properties[attr.obj]

        if env.ufunc_types:
            return NestedProperty(
                left=dtype,
                right=env.call('_resolve_property', prop.dtype, attr.name)
            )

        return env.call('_resolve_property', prop, attr.name)
    raise FieldNotInResource(dtype, property=attr.obj)


@ufunc.resolver(PropertyResolver, Object, Bind)
def _resolve_property(
    env: PropertyResolver,
    dtype: Object,
    attr: Bind
):
    if attr.name in dtype.properties:
        prop = dtype.properties[attr.name]

        if env.ufunc_types:
            return NestedProperty(
                left=dtype,
                right=prop.dtype
            )

        return prop
    else:
        raise FieldNotInResource(dtype, property=attr.name)


@ufunc.resolver(PropertyResolver, Array, (Bind, GetAttr))
def _resolve_property(env, dtype, attr):
    if env.ufunc_types:
        return NestedProperty(
            left=dtype,
            right=env.call('_resolve_property', dtype.items.dtype, attr)
        )

    return env.call('_resolve_property', dtype.items, attr)


@ufunc.resolver(PropertyResolver, File, Bind)
def _resolve_property(env, dtype, attr):
    return ReservedProperty(dtype, attr.name)


@ufunc.resolver(PropertyResolver, Text, Bind)
def _resolve_property(
    env: PropertyResolver,
    dtype: Text,
    bind: Bind
):
    if dtype.prop.name in env.model.properties:
        prop = env.model.properties[dtype.prop.name]
        if bind.name in prop.dtype.langs:
            return prop.dtype.langs[bind.name].dtype
        raise LangNotDeclared(dtype, lang=bind.name)


@ufunc.resolver(PropertyResolver, Inherit, Bind)
def _resolve_property(env, dtype, attr):
    return InheritForeignProperty(dtype.prop.model, attr.name, dtype.prop)


@ufunc.resolver(PropertyResolver, ForeignProperty, Property, object)
def _resolve_property(
    env: PropertyResolver,
    fpr: ForeignProperty,
    prop: Property,
    obj: object,
) -> ForeignProperty:
    return env.call('_resolve_property', fpr, prop.dtype, obj)


@ufunc.resolver(PropertyResolver, ForeignProperty, Ref, GetAttr)
def _resolve_property(
    env: PropertyResolver,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = fpr.push(prop)
    return env.call('_resolve_property', fpr, prop.dtype, attr.name)


@ufunc.resolver(PropertyResolver, ForeignProperty, Ref, Bind)
def _resolve_property(
    env: PropertyResolver,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return fpr.push(prop)


@ufunc.resolver(PropertyResolver, ForeignProperty, BackRef, GetAttr)
def _resolve_property(
    env: PropertyResolver,
    fpr: ForeignProperty,
    dtype: BackRef,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = fpr.push(prop)
    return env.call('_resolve_property', fpr, prop.dtype, attr.name)


@ufunc.resolver(PropertyResolver, ForeignProperty, BackRef, Bind)
def _resolve_property(
    env: PropertyResolver,
    fpr: ForeignProperty,
    dtype: BackRef,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return fpr.push(prop)


@ufunc.resolver(PropertyResolver, Bind)
def _resolve_property(
    env: PropertyResolver,
    bind: Bind
):
    return env.call('_resolve_property', bind.name)


@ufunc.resolver(PropertyResolver, GetAttr)
def _resolve_property(
    env: PropertyResolver,
    attr: GetAttr
):
    model = env.model
    place = str(attr)

    # Skip mapping if we don't need custom types
    if place in model.flatprops and not env.ufunc_types:
        return model.flatprops[place]

    prop = env.call('_resolve_property', attr.obj)
    return env.call('_resolve_property', prop, attr.name)


@ufunc.resolver(PropertyResolver, str)
def _resolve_property(
    env: PropertyResolver,
    prop: str
):
    if prop in env.model.flatprops:
        return env.model.flatprops.get(prop)

    raise PropertyNotFound(env.model, property=prop)


@ufunc.resolver(PropertyResolver, DataType)
def _resolve_property(
    env: PropertyResolver,
    dtype: DataType
):
    return dtype.prop


@ufunc.resolver(PropertyResolver, Property)
def _resolve_property(
    env: PropertyResolver,
    prop: Property
):
    return prop


@ufunc.resolver(PropertyResolver, ForeignProperty)
def _resolve_property(
    env: PropertyResolver,
    prop: ForeignProperty
):
    return prop


@ufunc.resolver(PropertyResolver, FuncProperty)
def _resolve_property(
    env: PropertyResolver,
    prop: FuncProperty
):
    return prop



from __future__ import annotations

from spinta.datasets.backends.sql.ufuncs.query.components import SqlQueryBuilder
from spinta.exceptions import GivenValueCountMissmatch, PropertyNotPartOfRefProps
from spinta.types.datatype import Ref, ExternalRef
from spinta.ufuncs.querybuilder.components import Selected


def select_ref_foreign_key_properties(
    env: SqlQueryBuilder,
    dtype: Ref,
    *,
    properties: list = None
) -> dict[str, Selected]:
    if properties is None:
        if dtype.prop.external and dtype.prop.external.prepare:
            properties = env(this=dtype.prop).resolve(dtype.prop.external.prepare)

    prep = {}
    given_mapping_count = len(properties) if properties else 1
    if len(dtype.refprops) != given_mapping_count:
        raise GivenValueCountMissmatch(dtype, given_count=given_mapping_count, expected_count=len(dtype.refprops))

    if properties is None:
        table = env.backend.get_table(env.model)
        column = env.backend.get_column(table, dtype.prop, select=True)
        refprop = dtype.refprops[0]
        prep[refprop.name] = Selected(item=env.add_column(column), prop=refprop)
    else:
        for i, prop in enumerate(dtype.refprops):
            sel = env.call('select', properties[i])
            if sel.prop is None:
                sel.prop = prop
            prep[prop.name] = sel
    return prep


def select_external_ref_foreign_key_properties(
    env: SqlQueryBuilder,
    dtype: ExternalRef,
    *,
    properties: list = None,
    target: str = None
) -> dict[str, Selected]:
    if properties is None:
        if dtype.prop.external and dtype.prop.external.prepare:
            properties = env(this=dtype.prop).resolve(dtype.prop.external.prepare)

    if target is not None:
        value = next((prop for prop in dtype.refprops if prop.name == target), None)
        if value is None:
            raise PropertyNotPartOfRefProps(dtype, prop=target, refprops=list(prop_.name for prop_ in dtype.refprops))
        target = value

    prep = {}
    given_mapping_count = len(properties) if properties else 1
    if len(dtype.refprops) != given_mapping_count:
        raise GivenValueCountMissmatch(dtype, given_count=given_mapping_count, expected_count=len(dtype.refprops))

    if properties is None:
        table = env.backend.get_table(env.model)
        column = env.backend.get_column(table, dtype.prop, select=True)
        refprop = dtype.refprops[0]
        prep[refprop.name] = Selected(item=env.add_column(column), prop=refprop)
    else:
        if target is not None:
            pos = dtype.refprops.index(target)
            sel = env.call('select', properties[pos])
            if sel.prop is None:
                sel.prop = target
            prep[target.name] = sel
        else:
            for i, prop in enumerate(dtype.refprops):
                sel = env.call('select', properties[i])
                if sel.prop is None:
                    sel.prop = prop
                prep[prop.name] = sel
    return prep



from __future__ import annotations

from spinta import commands
from spinta.components import Property
from spinta.types.datatype import Ref


@commands.resolve_property.register(Ref, str)
def resolve_property(dtype: Ref, prop: str) -> Property | None:
    return commands.resolve_property(dtype.model, prop)
